from abc import abstractmethod

import equinox as eqx
from jax import numpy as jnp
from jaxtyping import Array, ArrayLike, Float, Int


class Color(eqx.Module):
    """RGB in [0,1]. Helpers to convert to pygame-friendly formats."""

    r: Float[Array, ""]
    g: Float[Array, ""]
    b: Float[Array, ""]

    def __init__(
        self, r: Float[ArrayLike, ""], g: Float[ArrayLike, ""], b: Float[ArrayLike, ""]
    ):
        self.r = jnp.asarray(r)
        self.g = jnp.asarray(g)
        self.b = jnp.asarray(b)

    def to_tuple(
        self,
    ) -> tuple[Float[ArrayLike, ""], Float[ArrayLike, ""], Float[ArrayLike, ""]]:
        return (self.r, self.g, self.b)

    def to_rgb255(self) -> tuple[int, int, int]:
        r = int(float(jnp.clip(self.r, 0.0, 1.0)) * 255.0)
        g = int(float(jnp.clip(self.g, 0.0, 1.0)) * 255.0)
        b = int(float(jnp.clip(self.b, 0.0, 1.0)) * 255.0)
        return (r, g, b)

    def to_hex(self) -> str:
        r, g, b = self.to_rgb255()
        return f"#{r:02x}{g:02x}{b:02x}"

    @classmethod
    def from_hex(cls, hex_str: str) -> "Color":
        r = int(hex_str[1:3], 16) / 255.0
        g = int(hex_str[3:5], 16) / 255.0
        b = int(hex_str[5:7], 16) / 255.0
        return cls(r, g, b)

    @classmethod
    def from_rgb255(cls, r: int, g: int, b: int) -> "Color":
        return cls(r / 255.0, g / 255.0, b / 255.0)

    @classmethod
    def from_tuple(
        cls,
        rgb: tuple[Float[ArrayLike, ""], Float[ArrayLike, ""], Float[ArrayLike, ""]],
    ) -> "Color":
        return cls(*rgb)


WHITE = Color(1.0, 1.0, 1.0)
BLACK = Color(0.0, 0.0, 0.0)
GRAY = Color(0.78, 0.78, 0.78)
RED = Color(0.86, 0.24, 0.24)
GREEN = Color(0.24, 0.71, 0.29)
BLUE = Color(0.26, 0.53, 0.96)


class Transform(eqx.Module):
    """
    Affine mapping from world coords to screen pixels.

    Has a constant scale for all axis to avoid distortion.

    Attributes:
        width: Width of the screen in pixels.
        height: Height of the screen in pixels.
        scale: Scale factor from world units to pixels.
        offset: Offset to add after scaling.
        y_up: If True, the y-axis points up in world space.

    Args:
        width: Width of the screen in pixels.
        height: Height of the screen in pixels.
        scale: Scale factor from world units to pixels.
        offset: Offset to add after scaling.
        y_up: If True, the y-axis points up in world space.
    """

    width: Int[Array, ""]
    height: Int[Array, ""]

    scale: Float[Array, ""]
    offset: Float[Array, "2"]

    y_up: bool = True

    def __init__(
        self,
        width: Int[ArrayLike, ""],
        height: Int[ArrayLike, ""],
        scale: Float[ArrayLike, ""],
        offset: Float[ArrayLike, "2"],
        y_up: bool = True,
    ):
        self.width = jnp.asarray(width)
        self.height = jnp.asarray(height)
        self.scale = jnp.asarray(scale)
        self.offset = jnp.asarray(offset)
        self.y_up = y_up

    def world_to_px(self, point: Float[ArrayLike, "2"]) -> Int[Array, "2"]:
        """
        Return pixel coordinates corresponding to a point in world space.

        Args:
            point: A point in world space.

        Returns:
            The corresponding pixel coordinates.
        """
        pixel = jnp.asarray(point) * self.scale + self.offset

        if self.y_up:
            pixel = pixel.at[1].set(self.height - pixel[1])

        return pixel.astype(int)

    def scale_length(self, length: Float[ArrayLike, ""]) -> Float[Array, ""]:
        """
        Scale a length in world space to pixel space.

        Args:
            length: A length in world space.

        Returns:
            The corresponding length in pixel space.
        """
        return jnp.asarray(length * self.scale)


class AbstractRenderer(eqx.Module):
    """
    Base class for renderers.
    """

    @abstractmethod
    def open(self):
        """Open the rendering window."""

    @abstractmethod
    def close(self):
        """Close the rendering window."""

    @abstractmethod
    def draw(self):
        """Render the current scene to the window."""

    @abstractmethod
    def as_array(self) -> Float[ArrayLike, "H W 3"]:
        """Return the current frame as an RGB array."""


class Abstract2DRenderer(AbstractRenderer):
    """
    Base class for 2D renderers with vector drawing primitives.

    Attributes:
        transform: The transform from world coordinates to screen pixels.
    """

    transform: eqx.AbstractVar[Transform]

    @abstractmethod
    def draw_circle(
        self, center: Float[ArrayLike, "2"], radius: Float[ArrayLike, ""], color: Color
    ):
        """
        Draw a filled circle.

        Args:
            center: Center of the circle in world coordinates.
            radius: Radius of the circle in world units.
            color: Color of the circle.
        """

    @abstractmethod
    def draw_line(
        self,
        start: Float[ArrayLike, "2"],
        end: Float[ArrayLike, "2"],
        color: Color,
        width: Float[ArrayLike, ""] = 1,
    ):
        """
        Draw a line segment.

        Args:
            start: Start point of the line in world coordinates.
            end: End point of the line in world coordinates.
            color: Color of the line.
            width: Width of the line in pixels.
        """

    @abstractmethod
    def draw_rect(
        self,
        center: Float[ArrayLike, "2"],
        w: Float[ArrayLike, ""],
        h: Float[ArrayLike, ""],
        color: Color,
    ):
        """
        Draw a filled rectangle.

        Args:
            center: Center of the rectangle in world coordinates.
            w: Width of the rectangle in world units.
            h: Height of the rectangle in world units.
            color: Color of the rectangle.
        """

    @abstractmethod
    def draw_polygon(self, points: Float[ArrayLike, "num 2"], color: Color):
        """
        Draw a filled polygon.

        Args:
            points: Vertices of the polygon in world coordinates.
            color: Color of the polygon.
        """

    @abstractmethod
    def draw_text(
        self,
        center: Float[ArrayLike, "2"],
        text: str,
        color: Color,
        size: Float[ArrayLike, ""] = 12,
    ):
        """
        Draw text at a given position.

        Args:
            center: Center of the text in world coordinates.
            text: The text string to draw.
            color: Color of the text.
            size: Font size in points.
        """

    @abstractmethod
    def draw_polyline(
        self,
        points: Float[ArrayLike, "num 2"],
        color: Color,
    ):
        """
        Draw a polyline connecting a series of points.

        Args:
            points: Sequence of points in world coordinates.
            color: Color of the polyline.
        """

    @abstractmethod
    def draw_ellipse(
        self,
        center: Float[ArrayLike, "2"],
        w: Float[ArrayLike, ""],
        h: Float[ArrayLike, ""],
        color: Color,
    ):
        """
        Draw a filled ellipse.

        Args:
            center: Center of the ellipse in world coordinates.
            w: Width of the ellipse in world units.
            h: Height of the ellipse in world units.
            color: Color of the ellipse.
        """

    @abstractmethod
    def clear(self):
        """
        Clear the screen to a solid color.
        """


class Abstract3DRenderer(AbstractRenderer):
    """
    Base class for 3D renderers.
    """

    pass



from __future__ import annotations

from datetime import datetime

import optax
from jax import lax
from jax import numpy as jnp
from jaxtyping import Array, ArrayLike, Bool, Float, Int, Key, Scalar, ScalarLike
from tensorboardX import SummaryWriter

from lerax.env import AbstractEnvLike
from lerax.policy import AbstractPolicy
from lerax.utils import callback_with_numpy_wrapper

from .base_callback import (
    AbstractCallback,
    AbstractCallbackStepState,
    EmptyCallbackState,
    IterationContext,
    ResetContext,
    StepContext,
    TrainingContext,
)


class JITSummaryWriter:
    """
    A wrapper around `tensorboardX.SummaryWriter` with a JIT compatible interface.

    Attributes:
        summary_writer: The underlying SummaryWriter instance.

    Args:
        log_dir: Directory to save TensorBoard logs. If None, uses default.
    """

    summary_writer: SummaryWriter

    def __init__(self, log_dir: str | None = None):
        self.summary_writer = SummaryWriter(log_dir=log_dir)

    def add_scalar(
        self,
        tag: str,
        scalar_value: ScalarLike,
        global_step: Int[ArrayLike, ""] | None = None,
        walltime: Float[ArrayLike, ""] | None = None,
    ):
        """
        Add a scalar value to the summary writer.
        """
        callback_with_numpy_wrapper(self.summary_writer.add_scalar)(
            tag, scalar_value, global_step, walltime
        )

    def add_dict(
        self,
        prefix: str,
        scalars: dict[str, Scalar],
        *,
        global_step: Int[ArrayLike, ""] | None = None,
        walltime: Float[ArrayLike, ""] | None = None,
    ) -> None:
        """
        Log a dictionary of **scalar** values.
        """

        if prefix:
            scalars = {f"{prefix}/{k}": v for k, v in scalars.items()}

        for tag, value in scalars.items():
            self.add_scalar(tag, value, global_step=global_step, walltime=walltime)


class TensorBoardCallbackStepState(AbstractCallbackStepState):
    """
    State for TensorBoardCallback.

    Records cumulative episode returns and lengths, and the exponential moving
    average of them over episodes.

    Attributes:
        step: Current training step.
        episode_return: Cumulative return for the current episode.
        episode_length: Length of the current episode.
        episode_done: Boolean indicating if the current episode is done.
        average_return: Exponential moving average of episode returns.
        average_length: Exponential moving average of episode lengths.

    Args:
        step: Current training step.
        episode_return: Cumulative return for the current episode.
        episode_length: Length of the current episode.
        episode_done: Boolean indicating if the current episode is done.
        average_return: Exponential moving average of episode returns.
        average_length: Exponential moving average of episode lengths.
    """

    step: Int[Array, ""]

    episode_return: Float[Array, ""]
    episode_length: Int[Array, ""]
    episode_done: Bool[Array, ""]

    average_return: Float[Array, ""]
    average_length: Float[Array, ""]

    def __init__(
        self,
        step: Int[Array, ""],
        episode_return: Float[ArrayLike, ""],
        episode_length: Int[ArrayLike, ""],
        episode_done: Bool[ArrayLike, ""],
        average_return: Float[ArrayLike, ""],
        average_length: Float[ArrayLike, ""],
    ):
        self.step = jnp.asarray(step)
        self.episode_return = jnp.asarray(episode_return)
        self.episode_length = jnp.asarray(episode_length)
        self.episode_done = jnp.asarray(episode_done)
        self.average_return = jnp.asarray(average_return)
        self.average_length = jnp.asarray(average_length)

    @classmethod
    def initial(cls) -> TensorBoardCallbackStepState:
        return cls(
            jnp.array(0, dtype=int),
            jnp.array(0.0, dtype=float),
            jnp.array(0, dtype=int),
            jnp.array(False, dtype=bool),
            jnp.array(0.0, dtype=float),
            jnp.array(0.0, dtype=float),
        )

    def next(
        self, reward: Float[Array, ""], done: Bool[Array, ""], alpha: float
    ) -> TensorBoardCallbackStepState:
        average_return = lax.select(
            done,
            alpha * self.episode_return + (1.0 - alpha) * self.average_return,
            self.average_return,
        )
        average_length = lax.select(
            done,
            alpha * self.episode_length.astype(float)
            + (1.0 - alpha) * self.average_length,
            self.average_length,
        )

        return TensorBoardCallbackStepState(
            self.step + 1,
            self.episode_return * (1.0 - self.episode_done.astype(float)) + reward,
            self.episode_length * (1 - self.episode_done.astype(int)) + 1,
            done,
            average_return,
            average_length,
        )


class TensorBoardCallback(
    AbstractCallback[EmptyCallbackState, TensorBoardCallbackStepState]
):
    """
    Callback for recording training statistics to TensorBoard.

    Each training iteration, the following statistics are logged:
        - episode/return: The exponential moving average of episode returns.
        - episode/length: The exponential moving average of episode lengths.
        - train/:
            - learning_rate: The current learning rate.
            - ...: Any other statistics in the training log.

    Note:
        If the callback is instantiated inside a JIT-compiled function, it may
        not work correctly.

    Attributes:
        tb_writer: The TensorBoard summary writer.
        alpha: Smoothing factor for exponential moving averages.

    Args:
        name: Name for the TensorBoard log directory. If None, a name
            is generated based on the current time, environment name, and policy name.
        env: The environment being trained on. Used for naming if `name` is None.
        policy: The policy being trained. Used for naming if `name` is None.
        alpha: Smoothing factor for exponential moving averages.
    """

    tb_writer: JITSummaryWriter
    alpha: float

    def __init__(
        self,
        name: str | None = None,
        env: AbstractEnvLike | None = None,
        policy: AbstractPolicy | None = None,
        alpha: float = 0.9,
    ):
        log_dir = "logs/"
        time = datetime.now().strftime("%Y%m%d_%H%M%S")
        if name is None:
            if env is not None:
                if policy is not None:
                    name = f"{log_dir}{policy.name}_{env.name}_{time}"
                else:
                    name = f"{log_dir}{env.name}_{time}"
            else:
                if policy is not None:
                    name = f"{log_dir}{policy.name}_{time}"
                else:
                    name = f"{log_dir}training_{time}"

        self.tb_writer = JITSummaryWriter(log_dir=name)
        self.alpha = alpha

    def reset(self, ctx: ResetContext, *, key: Key) -> EmptyCallbackState:
        return EmptyCallbackState()

    def step_reset(
        self, ctx: ResetContext, *, key: Key
    ) -> TensorBoardCallbackStepState:
        return TensorBoardCallbackStepState.initial()

    def on_step(self, ctx: StepContext, *, key: Key) -> TensorBoardCallbackStepState:
        return ctx.state.next(ctx.reward, ctx.done, self.alpha)

    def on_iteration(self, ctx: IterationContext, *, key: Key) -> EmptyCallbackState:
        log = ctx.training_log
        opt_state = ctx.opt_state

        log["learning_rate"] = optax.tree_utils.tree_get(
            opt_state,
            "learning_rate",
            jnp.nan,
            filtering=lambda _, value: isinstance(value, jnp.ndarray),
        )

        step_state = ctx.step_state

        last_step = step_state.step.sum()

        self.tb_writer.add_dict("train", log, global_step=last_step)
        self.tb_writer.add_scalar(
            "episode/return", step_state.average_return.mean(), last_step
        )
        self.tb_writer.add_scalar(
            "episode/length", step_state.average_length.mean(), last_step
        )

        return ctx.state

    def on_training_start(
        self, ctx: TrainingContext, *, key: Key
    ) -> EmptyCallbackState:
        return ctx.state

    def on_training_end(self, ctx: TrainingContext, *, key: Key) -> EmptyCallbackState:
        return ctx.state

    def continue_training(self, ctx: IterationContext, *, key: Key) -> Bool[Array, ""]:
        return jnp.array(True)



from __future__ import annotations

from abc import abstractmethod

import equinox as eqx
from jax import lax
from jax import random as jr
from jaxtyping import Array, Float, Integer

from lerax.space import AbstractSpace, Discrete

from ..base_policy import (
    AbstractPolicyState,
    AbstractStatefulPolicy,
    AbstractStatefulWrapper,
    AbstractStatelessPolicy,
    NullStatefulPolicyState,
)


class AbstractStatelessQPolicy[ObsType](
    AbstractStatelessPolicy[Integer[Array, ""], ObsType]
):
    """
    Base class for stateless Q-learning policies.

    Q-learning policies map observations to actions by selecting the action
    with the highest Q-value, with epsilon-greedy exploration.

    Attributes:
        name: Name of the policy class.
        action_space: The action space of the environment.
        observation_space: The observation space of the environment.
        epsilon: The epsilon value for epsilon-greedy action selection.
    """

    name: eqx.AbstractClassVar[str]
    action_space: eqx.AbstractVar[Discrete]
    observation_space: eqx.AbstractVar[AbstractSpace[ObsType]]

    epsilon: eqx.AbstractVar[float]

    @abstractmethod
    def q_values(self, observation: ObsType) -> Float[Array, " actions"]:
        """
        Return Q-values for all actions given an observation.

        Args:
            observation: The current observation from the environment.

        Returns:
            Q-values for all actions.
        """

    def into_stateful[SelfType: AbstractStatelessQPolicy](
        self: SelfType,
    ) -> QStatefulWrapper[SelfType, ObsType]:
        return QStatefulWrapper(self)

    def __call__(
        self, observation: ObsType, *, key: Array | None = None
    ) -> Integer[Array, ""]:
        """
        Return the action for a given observation.

        Args:
            observation: The current observation from the environment.
            key: JAX PRNG key for stochastic action selection. If None, the
                action with the highest Q-value is always selected.

        Returns:
            The selected action.
        """
        q_vals = self.q_values(observation)

        if key is None or self.epsilon <= 0.0:
            return q_vals.argmax(axis=-1)
        else:
            epsilon_key, action_key = jr.split(key, 2)
            return lax.cond(
                jr.uniform(epsilon_key, shape=()) < self.epsilon,
                lambda: jr.randint(
                    action_key, shape=(), minval=0, maxval=self.action_space.n
                ),
                lambda: q_vals.argmax(axis=-1),
            )


class AbstractStatefulQPolicy[StateType: AbstractPolicyState, ObsType](
    AbstractStatefulPolicy[StateType, Integer[Array, ""], ObsType]
):
    """
    Base class for stateful Q-learning policies.

    Q-learning policies map observations and internal states to actions by
    selecting the action with the highest Q-value, with epsilon-greedy
    exploration.

    Attributes:
        name: Name of the policy class.
        action_space: The action space of the environment.
        observation_space: The observation space of the environment.
        epsilon: The epsilon value for epsilon-greedy action selection.
    """

    name: eqx.AbstractClassVar[str]
    action_space: eqx.AbstractVar[Discrete]
    observation_space: eqx.AbstractVar[AbstractSpace[ObsType]]

    epsilon: eqx.AbstractVar[float]

    @abstractmethod
    def q_values(
        self, state: StateType, observation: ObsType
    ) -> tuple[StateType, Float[Array, " actions"]]:
        """
        Return Q-values for all actions given an observation and state.

        Args:
            state: The current internal state of the policy.
            observation: The current observation from the environment.

        Returns:
            A tuple of the next internal state and the Q-values for all actions.
        """

    def __call__(
        self, state: StateType, observation: ObsType, *, key: Array | None = None
    ) -> tuple[StateType, Integer[Array, ""]]:
        """
        Return the next state and action for a given observation and state.

        Uses epsilon-greedy action selection.

        Args:
            state: The current internal state of the policy.
            observation: The current observation from the environment.
            key: JAX PRNG key for stochastic action selection. If None, the
                action with the highest Q-value is always selected.

        Returns:
            A tuple of the next internal state and the selected action.
        """
        state, q_vals = self.q_values(state, observation)

        if key is None or self.epsilon <= 0.0:
            return state, q_vals.argmax(axis=-1)
        else:
            epsilon_key, action_key = jr.split(key, 2)
            action = lax.cond(
                jr.uniform(epsilon_key, shape=()) < self.epsilon,
                lambda: jr.randint(
                    action_key, shape=(), minval=0, maxval=self.action_space.n
                ),
                lambda: q_vals.argmax(axis=-1),
            )
            return state, action


class QStatefulWrapper[PolicyType: AbstractStatelessQPolicy, ObsType](
    AbstractStatefulQPolicy[NullStatefulPolicyState, ObsType],
    AbstractStatefulWrapper[PolicyType, Integer[Array, ""], ObsType],
):
    policy: PolicyType

    def __init__(self, policy: PolicyType):
        self.policy = policy

    @property
    def epsilon(self) -> float:
        return self.policy.epsilon

    def q_values(
        self, state: NullStatefulPolicyState, observation: ObsType
    ) -> tuple[NullStatefulPolicyState, Float[Array, " actions"]]:
        return state, self.policy.q_values(observation)


type AbstractQPolicy = AbstractStatefulQPolicy | AbstractStatelessQPolicy



from __future__ import annotations

from typing import Any

from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, ArrayLike, Bool, Float, Int, Key

from .base_space import AbstractSpace
from .utils import try_cast


class MultiDiscrete(AbstractSpace[Int[Array, " n"]]):
    """
    Cartesian product of discrete spaces.

    Attributes:
        ns: The number of discrete values for each dimension.

    Args:
        ns: The number of discrete values for each dimension.
    """

    ns: tuple[int, ...]

    def __init__(self, ns: tuple[int, ...]):
        assert len(ns) > 0, "ns must be non-empty"
        assert all(n > 0 for n in ns), "all n must be positive"

        self.ns = ns

    @property
    def shape(self) -> tuple[int, ...]:
        return (len(self.ns),)

    def canonical(self) -> Int[Array, " n"]:
        return jnp.zeros(self.shape, dtype=int)

    def sample(self, key: Key) -> Int[Array, " n"]:
        return jr.randint(
            key,
            shape=self.shape,
            minval=0,
            maxval=jnp.array(self.ns, dtype=int),
            dtype=int,
        )

    def contains(self, x: Any) -> Bool[Array, ""]:
        x = try_cast(x)
        if x is None:
            return jnp.array(False)

        if x.shape != self.shape:
            return jnp.array(False)

        if ~jnp.array_equal(x, jnp.floor(x)):
            return jnp.array(False)

        return jnp.all(x < self.ns)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MultiDiscrete):
            return False

        return self.ns == other.ns

    def __repr__(self) -> str:
        return f"MultiDiscrete({self.ns})"

    def __hash__(self) -> int:
        return hash(self.ns)

    def flatten_sample(self, sample: Int[ArrayLike, " n"]) -> Float[Array, " n"]:
        return jnp.asarray(sample, dtype=float).ravel()

    @property
    def flat_size(self) -> int:
        return len(self.ns)



from __future__ import annotations

from typing import ClassVar

import diffrax
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, ArrayLike, Bool, Float, Int, Key

from lerax.render import (
    Abstract2DRenderer,
    AbstractRenderer,
    Color,
    PygameRenderer,
    Transform,
)
from lerax.space import Box, Discrete

from .base_classic_control import (
    AbstractClassicControlEnv,
    AbstractClassicControlEnvState,
)


class CartPoleState(AbstractClassicControlEnvState):
    y: Float[Array, "4"]
    t: Float[Array, ""]


class CartPole(
    AbstractClassicControlEnv[CartPoleState, Int[Array, ""], Float[Array, "4"]]
):
    """
    CartPole environment matching the [Gymnasium Cart Pole environment](https://gymnasium.farama.org/environments/classic_control/cart_pole/).

    Note:
        To achieve identical dynamics to Gymnasium set `solver=diffrax.Euler()`.

    ## Action Space

    The action space is discrete with two actions:

    - 0: Push cart to the left
    - 1: Push cart to the right

    The action applies a fixed magnitude force to the cart in the specified direction for the duration of the time step.

    ## Observation Space

    The observation space is a 4-dimensional continuous space representing the state of the cart and pole:

    | Index | Observation           | Min Value           | Max Value          |
    |-------|-----------------------|---------------------|--------------------|
    | 0     | Cart Position         | -4.8                | 4.8                |
    | 1     | Cart Velocity         | -Inf                | Inf                |
    | 2     | Pole Angle            | -24 deg (-0.418 rad)| 24 deg (0.418 rad) |
    | 3     | Pole Angular Velocity | -Inf                | Inf                |

    These values are double the termination thresholds to allow for some margin.
    These limits can be modified via the `theta_threshold_radians` and `x_threshold` parameters.

    ## Reward

    The reward is 1 for every step taken, including the termination step.

    ## Termination

    The episode terminates when:

    - The pole angle exceeds ±12 degrees from vertical.
    - The cart position exceeds ±2.4 units from the center.

    These values can be modified via the `theta_threshold_radians` and `x_threshold` parameters.

    Args:
        gravity: The gravity constant.
        cart_mass: The mass of the cart.
        pole_mass: The mass of the pole.
        half_length: The half-length of the pole.
        force_mag: The magnitude of the force applied to the cart.
        theta_threshold_radians: The angle threshold for terminating the episode.
        x_threshold: The position threshold for terminating the episode.
        dt: The time step for the simulation.
        solver: The differential equation solver used for simulating the dynamics.
        stepsize_controller: The step size controller for the solver.
    """

    name: ClassVar[str] = "CartPole"

    action_space: Discrete
    observation_space: Box

    gravity: Float[Array, ""]
    cart_mass: Float[Array, ""]
    pole_mass: Float[Array, ""]
    total_mass: Float[Array, ""]
    length: Float[Array, ""]
    polemass_length: Float[Array, ""]
    force_mag: Float[Array, ""]
    theta_threshold_radians: Float[Array, ""]
    x_threshold: Float[Array, ""]

    dt: Float[Array, ""]
    solver: diffrax.AbstractSolver
    dt0: Float[Array, ""] | None
    stepsize_controller: diffrax.AbstractStepSizeController

    def __init__(
        self,
        *,
        gravity: Float[ArrayLike, ""] = 9.8,
        cart_mass: Float[ArrayLike, ""] = 1.0,
        pole_mass: Float[ArrayLike, ""] = 0.1,
        half_length: Float[ArrayLike, ""] = 0.5,
        force_mag: Float[ArrayLike, ""] = 10.0,
        theta_threshold_radians: Float[ArrayLike, ""] = 12 * 2 * jnp.pi / 360,
        x_threshold: Float[ArrayLike, ""] = 2.4,
        dt: Float[ArrayLike, ""] = 0.02,
        solver: diffrax.AbstractSolver | None = None,
        stepsize_controller: diffrax.AbstractStepSizeController | None = None,
    ):
        self.gravity = jnp.array(gravity)
        self.cart_mass = jnp.array(cart_mass)
        self.pole_mass = jnp.array(pole_mass)
        self.total_mass = self.pole_mass + self.cart_mass
        self.length = jnp.array(half_length)
        self.polemass_length = self.pole_mass * self.length
        self.force_mag = jnp.array(force_mag)

        self.dt = jnp.array(dt)
        self.solver = solver or diffrax.Tsit5()
        is_adaptive = isinstance(self.solver, diffrax.AbstractAdaptiveSolver)
        self.dt0 = None if is_adaptive else self.dt
        if stepsize_controller is None:
            if is_adaptive:
                self.stepsize_controller = diffrax.PIDController(rtol=1e-5, atol=1e-5)
            else:
                self.stepsize_controller = diffrax.ConstantStepSize()
        else:
            self.stepsize_controller = stepsize_controller

        self.theta_threshold_radians = jnp.array(theta_threshold_radians)
        self.x_threshold = jnp.array(x_threshold)

        self.action_space = Discrete(2)
        high = jnp.array(
            [
                self.x_threshold * 2,
                jnp.inf,
                self.theta_threshold_radians * 2,
                jnp.inf,
            ],
        )
        self.observation_space = Box(-high, high)

    def initial(self, *, key: Key) -> CartPoleState:
        return CartPoleState(
            y=jr.uniform(key, (4,), minval=-0.05, maxval=0.05), t=jnp.array(0.0)
        )

    def dynamics(
        self, t: Float[Array, ""], y: Float[Array, "4"], action: Int[Array, ""]
    ) -> Float[Array, "4"]:
        _, x_dot, theta, theta_dot = y
        force = (action * 2 - 1) * self.force_mag

        temp = (
            force + self.polemass_length * theta_dot**2 * jnp.sin(theta)
        ) / self.total_mass
        theta_dd = (self.gravity * jnp.sin(theta) - jnp.cos(theta) * temp) / (
            self.length
            * (4.0 / 3.0 - self.pole_mass * (jnp.cos(theta) ** 2) / self.total_mass)
        )
        x_dd = temp - self.polemass_length * theta_dd * jnp.cos(theta) / self.total_mass

        return jnp.array([x_dot, x_dd, theta_dot, theta_dd])

    def clip(self, y: Float[Array, "4"]) -> Float[Array, "4"]:
        return y

    def observation(self, state: CartPoleState, *, key: Key) -> Float[Array, "4"]:
        return state.y

    def reward(
        self,
        state: CartPoleState,
        action: Int[Array, ""],
        next_state: CartPoleState,
        *,
        key: Key,
    ) -> Float[Array, ""]:
        return jnp.array(1.0)

    def terminal(self, state: CartPoleState, *, key: Key) -> Bool[Array, ""]:
        x, theta = state.y[0], state.y[2]
        within_x = (x >= -self.x_threshold) & (x <= self.x_threshold)
        within_theta = (theta >= -self.theta_threshold_radians) & (
            theta <= self.theta_threshold_radians
        )
        return ~(within_x & within_theta)

    def render(self, state: CartPoleState, renderer: AbstractRenderer):
        if not isinstance(renderer, Abstract2DRenderer):
            raise TypeError("CartPole environment requires a 2D renderer.")

        x, theta = state.y[0], state.y[2]

        renderer.clear()

        # Ground
        renderer.draw_line(
            start=jnp.array((-10.0, 0.0)),
            end=jnp.array((10.0, 0.0)),
            color=Color(0.0, 0.0, 0.0),
            width=0.01,
        )
        # Cart
        cart_w, cart_h = 0.3, 0.15
        cart_col = Color(0.0, 0.0, 0.0)
        renderer.draw_rect(jnp.array((x, 0.0)), w=cart_w, h=cart_h, color=cart_col)
        # Pole
        pole_start = jnp.asarray((x, cart_h / 4))
        pole_end = pole_start + self.length * jnp.asarray(
            [jnp.sin(theta), jnp.cos(theta)]
        )
        pole_col = Color(0.8, 0.6, 0.4)
        renderer.draw_line(pole_start, pole_end, color=pole_col, width=0.05)
        # Pole Hinge
        hinge_r = 0.025
        hinge_col = Color(0.5, 0.5, 0.5)
        renderer.draw_circle(pole_start, radius=hinge_r, color=hinge_col)

        renderer.draw()

    def default_renderer(self) -> Abstract2DRenderer:
        width, height = 800, 450
        transform = Transform(
            scale=200.0,
            offset=jnp.array([width / 2, height * 0.1]),
            width=width,
            height=height,
            y_up=True,
        )
        return PygameRenderer(
            width=width,
            height=height,
            background_color=Color(1.0, 1.0, 1.0),
            transform=transform,
        )



from __future__ import annotations

from functools import partial, wraps
from pathlib import Path
from typing import Any, Callable, Sequence

import equinox as eqx
import jax
import numpy as np
from jax import lax
from jax import numpy as jnp


def filter_scan(
    f,
    init,
    xs=None,
    length=None,
    reverse: bool = False,
    unroll: int | bool = 1,
    _split_transpose: bool = False,
):
    """
    An easier to use version of `lax.scan`. All JAX and Numpy arrays are
    traced, and only non-array parts of the carry are static.

    Args:
        f: a Python function to be scanned of type ``c -> a -> (c, b)``, meaning
        that ``f`` accepts two arguments where the first is a value of the loop
        carry and the second is a slice of ``xs`` along its leading axis, and that
        ``f`` returns a pair where the first element represents a new value for
        the loop carry and the second represents a slice of the output.
        init: an initial loop carry value of type ``c``, which can be a scalar,
        array, or any pytree (nested Python tuple/list/dict) thereof, representing
        the initial loop carry value. This value must have the same structure as
        the first element of the pair returned by ``f``.
        xs: the value of type ``[a]`` over which to scan along the leading axis,
        where ``[a]`` can be an array or any pytree (nested Python
        tuple/list/dict) thereof with consistent leading axis sizes.
        length: optional integer specifying the number of loop iterations, which
        must agree with the sizes of leading axes of the arrays in ``xs`` (but can
        be used to perform scans where no input ``xs`` are needed).
        reverse: optional boolean specifying whether to run the scan iteration
        forward (the default) or in reverse, equivalent to reversing the leading
        axes of the arrays in both ``xs`` and in ``ys``.
        unroll: optional non-negative int or bool specifying, in the underlying
        operation of the scan primitive, how many scan iterations to unroll within
        a single iteration of a loop. If an integer is provided, it determines how
        many unrolled loop iterations to run within a single rolled iteration of
        the loop. `unroll=0` unrolls the entire loop.
        If a boolean is provided, it will determine if the loop is
        completely unrolled (i.e. `unroll=True`) or left completely rolled (i.e.
        `unroll=False`).
        _split_transpose: experimental optional bool specifying whether to further
        split the transpose into a scan (computing activation gradients), and a
        map (computing gradients corresponding to the array arguments). Enabling
        this may increase memory requirements, and so is an experimental feature
        that may evolve or even be rolled back.

    Returns:
        A pair of type ``(c, [b])`` where the first element represents the final
        loop carry value and the second element represents the stacked outputs of
        the second output of ``f`` when scanned over the leading axis of the inputs.
    """
    init_arr, static = eqx.partition(init, eqx.is_array)

    def _f(carry_arr, x):
        carry = eqx.combine(carry_arr, static)
        carry, y = f(carry, x)
        new_carry_arr, new_static = eqx.partition(carry, eqx.is_array)
        assert eqx.tree_equal(
            static, new_static
        ), "Non-array carry of filter_scan must not change."
        return new_carry_arr, y

    carry_arr, ys = lax.scan(
        f=_f,
        init=init_arr,
        xs=xs,
        length=length,
        reverse=reverse,
        unroll=unroll,
        _split_transpose=_split_transpose,
    )
    return eqx.combine(carry_arr, static), ys


def callback_wrapper[**InType](
    func: Callable[InType, Any], ordered: bool = False
) -> Callable[InType, None]:
    """
    Return a JIT‑safe version of *func*.

    Wraps *func* in a `jax.debug.callback` so that it can be used inside JIT‑compiled
    code.

    Args:
        func: The callback function to wrap.
        ordered: Whether to enforce ordered execution of callbacks.

    Returns:
        A wrapped version of *func* that is JIT-safe.
    """

    def _callback(*args: InType.args, **kwargs: InType.kwargs) -> None:
        func(*args, **kwargs)

    @wraps(func)
    def wrapped(*args: InType.args, **kwargs: InType.kwargs) -> None:
        jax.debug.callback(_callback, *args, **kwargs, ordered=ordered)

    return wrapped


def callback_with_numpy_wrapper(
    func: Callable[..., Any], ordered: bool = False
) -> Callable[..., None]:
    """
    Like `debug_wrapper` but converts every jax.Array/`jnp.ndarray` argument
    to a plain `numpy.ndarray` before calling *func*.

    It is impossible with Python's current type system to express the
    transformation so parameter information is lost.

    Args:
        func: The callback function to wrap.
        ordered: Whether to enforce ordered execution of callbacks.

    Returns:
        A wrapped version of *func* that converts array arguments to numpy
        arrays and is JIT-safe.
    """

    @partial(callback_wrapper, ordered=ordered)
    @wraps(func)
    def wrapped(*args, **kwargs) -> None:
        args, kwargs = jax.tree.map(
            lambda x: np.asarray(x) if isinstance(x, jnp.ndarray) else x, (args, kwargs)
        )
        func(*args, **kwargs)

    return wrapped


def callback_with_list_wrapper(
    func: Callable[..., Any], ordered: bool = False
) -> Callable[..., None]:
    """
    Like `debug_wrapper` but converts every jax.Array/`jnp.ndarray` argument
    to a plain list before calling *func*.

    It is impossible with Python's current type system to express the
    transformation so parameter information is lost.

    Args:
        func: The callback function to wrap.
        ordered: Whether to enforce ordered execution of callbacks.

    Returns:
        A wrapped version of *func* that converts array arguments to lists and
        is JIT-safe.
    """

    @partial(callback_wrapper, ordered=ordered)
    @wraps(func)
    def wrapped(*args, **kwargs) -> None:
        args, kwargs = jax.tree.map(
            lambda x: (
                np.asarray(x).tolist()
                if isinstance(x, (jnp.ndarray, np.ndarray))
                else x
            ),
            (args, kwargs),
        )
        func(*args, **kwargs)

    return wrapped


print_callback = callback_with_list_wrapper(print, ordered=True)


def unstack_pytree[T](tree: T, *, axis: int = 0) -> Sequence[T]:
    """
    Split a stacked pytree along `axis` into a tuple of pytrees with the same
    structure.

    Args:
        tree: A pytree with array leaves stacked along `axis`.
        axis: The axis along which to unstack the arrays.

    Returns:
        A sequence of pytrees with the same structure, each corresponding to one
        slice along `axis`.
    """

    times = jnp.array(jax.tree.leaves(jax.tree.map(lambda x: x.shape[axis], tree)))
    tree = eqx.error_if(
        tree,
        ~times == times[0],
        "All leaves must have the same size along the specified axis.",
    )

    outer_structure = jax.tree.structure(tree)
    unstacked = jax.tree.map(partial(jnp.unstack, axis=axis), tree)
    transposed = jax.tree.transpose(outer_structure, None, unstacked)
    return transposed


class Serializable(eqx.Module):
    @callback_wrapper
    def serialize(
        self,
        path: str | Path,
        no_suffix: bool = False,
    ) -> None:
        """
        Serialize the model to the specified path.

        Args:
            path: The path to serialize to.
            no_suffix: If True, do not append the ".eqx" suffix
        """
        path = Path(path)
        if not path.parent.exists():
            path.parent.mkdir(parents=True, exist_ok=True)
        if path.suffix != ".eqx" and not no_suffix:
            path = path.with_suffix(".eqx")

        eqx.tree_serialise_leaves(path, self)

    @classmethod
    def deserialize[**Params, ClassType](
        cls: Callable[Params, ClassType],
        path: str | Path,
        *args: Params.args,
        **kwargs: Params.kwargs,
    ) -> ClassType:
        """
        Deserialize the model from the specified path.
        Must provide any additional arguments required by the class constructor.

        Args:
            path: The path to deserialize from.
            *args: Additional arguments to pass to the class constructor
            **kwargs: Additional keyword arguments to pass to the class constructor

        Returns:
            The deserialized model.
        """
        return eqx.tree_deserialise_leaves(
            path, eqx.filter_eval_shape(cls, *args, **kwargs)
        )



from __future__ import annotations

from typing import ClassVar

import diffrax
import equinox as eqx
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, Float, Integer, Key, Real

from lerax.env import AbstractEnvLike, AbstractEnvLikeState
from lerax.model import MLP, ActionLayer, MLPNeuralCDE, NCDEState
from lerax.space import AbstractSpace

from .base_actor_critic import (
    AbstractPolicyState,
    AbstractStatefulActorCriticPolicy,
)


class NCDEPolicyState(AbstractPolicyState):
    t: Float[Array, ""]
    cde: NCDEState

    def __init__(self, *, t: Float[Array, ""] = jnp.array(0.0), cde: NCDEState):
        self.t = jnp.asarray(t, dtype=float)
        self.cde = cde


class NCDEActorCriticPolicy[
    ActType: (Float[Array, " dims"], Integer[Array, ""]),
    ObsType: Real[Array, "..."],
](AbstractStatefulActorCriticPolicy[NCDEPolicyState, ActType, ObsType]):
    """
    Actor–critic with a shared MLPNeuralCDE encoder and MLP heads.

    Acts by encoding observations with a Neural CDE, then passing the
    encoded features to separate MLPs to produce action distributions and
    value estimates.

    Attributes:
        name: Name of the policy class.
        action_space: The action space of the environment.
        observation_space: The observation space of the environment.
        encoder: Neural CDE to encode observations into features.
        value_head: MLP to produce value estimates from features.
        action_head: MLP to produce action distributions from features.

    Args:
        env: The environment to create the policy for.
        solver: Diffrax solver to use for the Neural CDE.
        feature_size: Size of the feature representation.
        latent_size: Size of the latent state in the Neural CDE.
        field_width: Width of the hidden layers in the Neural CDE vector field.
        field_depth: Depth of the hidden layers in the Neural CDE vector field.
        initial_width: Width of the hidden layers in the Neural CDE initial network.
        initial_depth: Depth of the hidden layers in the Neural CDE initial network.
        value_width: Width of the hidden layers in the value head.
        value_depth: Depth of the hidden layers in the value head.
        action_width: Width of the hidden layers in the action head.
        action_depth: Depth of the hidden layers in the action head.
        history_length: Number of past observations to condition on.
        dt: Time step between observations for the Neural CDE.
        log_std_init: Initial log standard deviation for continuous action spaces.
        key: JAX PRNG key for parameter initialization.
    """

    name: ClassVar[str] = "NCDEActorCriticPolicy"

    action_space: AbstractSpace[ActType]
    observation_space: AbstractSpace[ObsType]

    encoder: MLPNeuralCDE
    value_head: MLP
    action_head: ActionLayer

    dt: float = eqx.field(static=True)

    def __init__[StateType: AbstractEnvLikeState](
        self,
        env: AbstractEnvLike[StateType, ActType, ObsType],
        *,
        solver: diffrax.AbstractSolver | None = None,
        feature_size: int = 4,
        latent_size: int = 4,
        field_width: int = 8,
        field_depth: int = 1,
        initial_width: int = 16,
        initial_depth: int = 1,
        value_width: int = 16,
        value_depth: int = 1,
        action_width: int = 16,
        action_depth: int = 1,
        history_length: int = 4,
        dt: float = 1.0,
        log_std_init: float = 0.0,
        key: Key,
    ):
        self.action_space = env.action_space
        self.observation_space = env.observation_space
        self.dt = float(dt)

        enc_key, val_key, act_key = jr.split(key, 3)

        self.encoder = MLPNeuralCDE(
            in_size=self.observation_space.flat_size,
            latent_size=latent_size,
            solver=solver,
            field_width=field_width,
            field_depth=field_depth,
            initial_width=initial_width,
            initial_depth=initial_depth,
            time_in_input=False,
            history_length=history_length,
            key=enc_key,
        )

        self.value_head = MLP(
            in_size=latent_size,
            out_size="scalar",
            width_size=value_width,
            depth=value_depth,
            key=val_key,
        )

        self.action_head = ActionLayer(
            self.action_space,
            feature_size,
            action_width,
            action_depth,
            key=act_key,
            log_std_init=log_std_init,
        )

    def _step_encoder(
        self, state: NCDEPolicyState, obs: ObsType
    ) -> tuple[NCDEPolicyState, Float[Array, " feat"]]:
        t_next = state.t + self.dt
        cde_state, y = self.encoder(
            state.cde, t_next, self.observation_space.flatten_sample(obs)
        )
        return NCDEPolicyState(t=t_next, cde=cde_state), y

    def reset(self, *, key: Key) -> NCDEPolicyState:
        return NCDEPolicyState(t=jnp.array(0.0), cde=self.encoder.reset())

    def __call__(
        self, state: NCDEPolicyState, observation: ObsType, *, key: Key | None = None
    ) -> tuple[NCDEPolicyState, ActType]:
        state, features = self._step_encoder(state, observation)
        features = jnp.tanh(features)

        action_dist = self.action_head(features)

        if key is None:
            action = action_dist.mode()
        else:
            action = action_dist.sample(key)

        return state, action

    def action_and_value(
        self, state: NCDEPolicyState, observation: ObsType, *, key: Key
    ) -> tuple[NCDEPolicyState, ActType, Float[Array, ""], Float[Array, ""]]:
        state, features = self._step_encoder(state, observation)
        features = jnp.tanh(features)

        value = self.value_head(features)

        action_dist = self.action_head(features)
        action, log_prob = action_dist.sample_and_log_prob(key)

        return state, action, value, log_prob.sum().squeeze()

    def evaluate_action(
        self, state: NCDEPolicyState, observation: ObsType, action: ActType
    ) -> tuple[NCDEPolicyState, Float[Array, ""], Float[Array, ""], Float[Array, ""]]:
        state, features = self._step_encoder(state, observation)
        features = jnp.tanh(features)

        action_dist = self.action_head(features)
        value = self.value_head(features)
        log_prob = action_dist.log_prob(action)

        eqx.debug.breakpoint_if(~jnp.isfinite(log_prob))

        try:
            entropy = action_dist.entropy().squeeze()
        except NotImplementedError:
            entropy = -log_prob.mean().squeeze()

        return state, value, log_prob.sum().squeeze(), entropy

    def value(
        self, state: NCDEPolicyState, observation: ObsType
    ) -> tuple[NCDEPolicyState, Float[Array, ""]]:
        state, feats = self._step_encoder(state, observation)
        return state, self.value_head(feats)



from __future__ import annotations

from abc import abstractmethod

import equinox as eqx
from jaxtyping import Array, Float, Key

from lerax.space import AbstractSpace

from ..base_policy import (
    AbstractPolicyState,
    AbstractStatefulPolicy,
    AbstractStatefulWrapper,
    AbstractStatelessPolicy,
    NullStatefulPolicyState,
)


class AbstractStatelessActorCriticPolicy[ActType, ObsType](
    AbstractStatelessPolicy[ActType, ObsType]
):
    """
    Base class for stateless actor-critic policies.

    Actor-critic policies map observations to actions and values.

    Attributes:
        name: The name of the policy.
        action_space: The action space of the policy.
        observation_space: The observation space of the policy.
    """

    name: eqx.AbstractClassVar[str]

    action_space: eqx.AbstractVar[AbstractSpace[ActType]]
    observation_space: eqx.AbstractVar[AbstractSpace[ObsType]]

    @abstractmethod
    def action_and_value(
        self, observation: ObsType, *, key: Key
    ) -> tuple[ActType, Float[Array, ""], Float[Array, ""]]:
        """
        Get an action and value from an observation.

        Args:
            observation: The observation to get the action and value for.
            key: A JAX PRNG key.

        Returns:
            action: The action to take.
            value: The value of the observation.
            log_prob: The log probability of the action.
        """

    @abstractmethod
    def evaluate_action(
        self, observation: ObsType, action: ActType
    ) -> tuple[Float[Array, ""], Float[Array, ""], Float[Array, ""]]:
        """
        Evaluate an action given an observation.

        Args:
            observation: The observation to evaluate the action for.
            action: The action to evaluate.

        Returns:
            value: The value of the observation.
            log_prob: The log probability of the action.
            entropy: The entropy of the action distribution.
        """

    @abstractmethod
    def value(self, observation: ObsType) -> Float[Array, ""]:
        """
        Get the value of an observation.

        Args:
            observation: The observation to get the value for.

        Returns:
            value: The value of the observation.
        """

    def into_stateful[SelfType: AbstractStatelessActorCriticPolicy](
        self: SelfType,
    ) -> ActorCriticStatefulWrapper[SelfType, ActType, ObsType]:
        """
        Transform this stateless policy into a stateful one.

        Returns:
            A stateful wrapper around this policy.
        """
        return ActorCriticStatefulWrapper(self)


class AbstractStatefulActorCriticPolicy[
    StateType: AbstractPolicyState, ActType, ObsType
](AbstractStatefulPolicy[StateType, ActType, ObsType]):
    """
    Base class for stateful actor-critic policies.

    Actor-critic policies map observations and internal states to actions, values, and new internal states.

    Attributes:
        name: The name of the policy.
        action_space: The action space of the policy.
        observation_space: The observation space of the policy.
    """

    name: eqx.AbstractClassVar[str]

    action_space: eqx.AbstractVar[AbstractSpace[ActType]]
    observation_space: eqx.AbstractVar[AbstractSpace[ObsType]]

    @abstractmethod
    def action_and_value(
        self, state: StateType, observation: ObsType, *, key: Key
    ) -> tuple[StateType, ActType, Float[Array, ""], Float[Array, ""]]:
        """
        Get an action and value from an observation.

        Args:
            state: The current policy state.
            observation: The observation to get the action and value for.
            key: A JAX PRNG key.

        Returns:
            new_state: The new policy state.
            action: The action to take.
            value: The value of the observation.
            log_prob: The log probability of the action.
        """

    @abstractmethod
    def evaluate_action(
        self, state: StateType, observation: ObsType, action: ActType
    ) -> tuple[StateType, Float[Array, ""], Float[Array, ""], Float[Array, ""]]:
        """
        Evaluate an action given an observation.

        Args:
            state: The current policy state.
            observation: The observation to evaluate the action for.
            action: The action to evaluate.

        Returns:
            new_state: The new policy state.
            value: The value of the observation.
            log_prob: The log probability of the action.
            entropy: The entropy of the action distribution.
        """

    @abstractmethod
    def value(
        self, state: StateType, observation: ObsType
    ) -> tuple[StateType, Float[Array, ""]]:
        """
        Get the value of an observation.

        Args:
            state: The current policy state.
            observation: The observation to get the value for.

        Returns:
            new_state: The new policy state.
            value: The value of the observation.
        """


class ActorCriticStatefulWrapper[
    PolicyType: AbstractStatelessActorCriticPolicy,
    ActType,
    ObsType,
](
    AbstractStatefulActorCriticPolicy[NullStatefulPolicyState, ActType, ObsType],
    AbstractStatefulWrapper[PolicyType, ActType, ObsType],
):
    policy: PolicyType

    def __init__(self, policy: PolicyType):
        self.policy = policy

    def action_and_value(
        self, state: NullStatefulPolicyState, observation: ObsType, *, key: Key
    ) -> tuple[NullStatefulPolicyState, ActType, Float[Array, ""], Float[Array, ""]]:
        return state, *self.policy.action_and_value(observation, key=key)

    def evaluate_action(
        self, state: NullStatefulPolicyState, observation: ObsType, action: ActType
    ) -> tuple[
        NullStatefulPolicyState, Float[Array, ""], Float[Array, ""], Float[Array, ""]
    ]:
        return state, *self.policy.evaluate_action(observation, action)

    def value(
        self, state: NullStatefulPolicyState, observation: ObsType
    ) -> tuple[NullStatefulPolicyState, Float[Array, ""]]:
        return state, self.policy.value(observation)


type AbstractActorCriticPolicy = AbstractStatefulActorCriticPolicy | AbstractStatelessActorCriticPolicy



from __future__ import annotations

from abc import abstractmethod

import equinox as eqx
import jax
import optax
from jax import lax
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, Int, Key, Scalar

from lerax.buffer import ReplayBuffer
from lerax.callback import (
    AbstractCallback,
    AbstractCallbackState,
    AbstractCallbackStepState,
    IterationContext,
    ResetContext,
    StepContext,
)
from lerax.env import AbstractEnvLike, AbstractEnvLikeState
from lerax.policy import AbstractPolicyState, AbstractStatefulPolicy
from lerax.space import Box
from lerax.utils import filter_scan

from .base_algorithm import AbstractAlgorithm, AbstractAlgorithmState, AbstractStepState


class OffPolicyStepState[PolicyType: AbstractStatefulPolicy](AbstractStepState):
    """
    State object for off-policy algorithms steps.

    Attributes:
        env_state: The state of the environment.
        policy_state: The state of the policy.
        callback_state: The state of the callback for this step.
        buffer: The replay buffer storing experience.
    """

    env_state: AbstractEnvLikeState
    policy_state: AbstractPolicyState
    callback_state: AbstractCallbackStepState
    buffer: ReplayBuffer

    @classmethod
    def initial(
        cls,
        size: int,
        env: AbstractEnvLike,
        policy: PolicyType,
        callback: AbstractCallback,
        key: Key,
    ) -> OffPolicyStepState[PolicyType]:
        """
        Initialize the off-policy step state.

        Resets the environment and policy, initializes the callback state,
        and creates an empty replay buffer.

        Args:
            size: The size of the replay buffer.
            env: The environment to initialize.
            policy: The policy to initialize.
            callback: The callback to initialize.
            key: A JAX PRNG key.

        Returns:
            The initialized step state.
        """
        env_key, policy_key = jr.split(key, 2)
        env_state = env.initial(key=env_key)
        policy_state = policy.reset(key=policy_key)

        callback_state = callback.step_reset(ResetContext(locals()), key=key)

        buffer = ReplayBuffer(
            size, env.observation_space, env.action_space, policy_state
        )
        return cls(env_state, policy_state, callback_state, buffer)


class OffPolicyState[PolicyType: AbstractStatefulPolicy](
    AbstractAlgorithmState[PolicyType]
):
    """
    State for off-policy algorithms.

    Attributes:
        iteration_count: The current iteration count.
        step_state: The current step state.
        env: The environment being used.
        policy: The policy being trained.
        opt_state: The optimizer state.
        callback_state: The state of the callback for this iteration.
    """

    iteration_count: Int[Array, ""]
    step_state: OffPolicyStepState[PolicyType]
    env: AbstractEnvLike
    policy: PolicyType
    opt_state: optax.OptState
    callback_state: AbstractCallbackState


class AbstractOffPolicyAlgorithm[PolicyType: AbstractStatefulPolicy](
    AbstractAlgorithm[PolicyType, OffPolicyState[PolicyType]]
):
    """
    Base class for off-policy algorithms.

    Generates experience using a policy and environment, stores it in a replay buffer,
    and trains the policy using samples from the replay buffer.

    Attributes:
        optimizer: The optimizer used for training the policy.
        buffer_size: The size of the replay buffer.
        gamma: The discount factor for future rewards.
        learning_starts: The number of initial steps to collect before training.
        num_envs: The number of parallel environments.
        batch_size: The batch size for training.
    """

    optimizer: eqx.AbstractVar[optax.GradientTransformation]

    buffer_size: eqx.AbstractVar[int]
    gamma: eqx.AbstractVar[float]
    learning_starts: eqx.AbstractVar[int]

    num_envs: eqx.AbstractVar[int]
    num_steps: eqx.AbstractVar[int]
    batch_size: eqx.AbstractVar[int]

    def num_iterations(self, total_timesteps: int) -> int:
        return total_timesteps // (self.num_envs * self.num_steps)

    @abstractmethod
    def per_step(
        self, step_state: OffPolicyStepState[PolicyType]
    ) -> OffPolicyStepState[PolicyType]:
        """Process the step carry after each step."""

    def step(
        self,
        env: AbstractEnvLike,
        policy: PolicyType,
        state: OffPolicyStepState[PolicyType],
        *,
        key: Key,
        callback: AbstractCallback,
    ) -> OffPolicyStepState[PolicyType]:
        (
            action_key,
            transition_key,
            observation_key,
            reward_key,
            terminal_key,
            next_observation_key,
            env_reset_key,
            policy_reset_key,
            callback_key,
        ) = jr.split(key, 9)

        observation = env.observation(state.env_state, key=observation_key)
        policy_state, action = policy(state.policy_state, observation, key=action_key)

        if isinstance(env.action_space, Box):
            clipped_action = jnp.clip(
                action,
                env.action_space.low,
                env.action_space.high,
            )
        else:
            clipped_action = action

        next_env_state = env.transition(
            state.env_state, clipped_action, key=transition_key
        )

        reward = env.reward(state.env_state, action, next_env_state, key=reward_key)
        termination = env.terminal(next_env_state, key=terminal_key)
        truncation = env.truncate(next_env_state)
        done = termination | truncation
        timeout = truncation & ~termination
        next_observation = env.observation(next_env_state, key=next_observation_key)

        next_env_state = lax.cond(
            done, lambda: env.initial(key=env_reset_key), lambda: next_env_state
        )

        next_policy_state = lax.cond(
            done, lambda: policy.reset(key=policy_reset_key), lambda: policy_state
        )

        replay_buffer = state.buffer.add(
            observation,
            next_observation,
            action,
            reward,
            done,
            timeout,
            state.policy_state,
            policy_state,
        )

        callback_state = callback.on_step(
            StepContext(state.callback_state, env, policy, done, reward, locals()),
            key=callback_key,
        )

        return OffPolicyStepState(
            next_env_state, next_policy_state, callback_state, replay_buffer
        )

    def collect_learning_starts(
        self,
        env: AbstractEnvLike,
        policy: PolicyType,
        step_state: OffPolicyStepState[PolicyType],
        callback: AbstractCallback,
        key: Key,
    ) -> OffPolicyStepState[PolicyType]:
        def scan_step(
            carry: OffPolicyStepState, key: Key
        ) -> tuple[OffPolicyStepState, None]:
            carry = self.step(env, policy, carry, key=key, callback=callback)
            return carry, None

        step_state, _ = filter_scan(
            scan_step, step_state, jr.split(key, self.learning_starts)
        )

        return step_state

    def collect_rollout(
        self,
        env: AbstractEnvLike,
        policy: PolicyType,
        step_state: OffPolicyStepState[PolicyType],
        callback: AbstractCallback,
        key: Key,
    ) -> OffPolicyStepState[PolicyType]:
        def scan_step(
            carry: OffPolicyStepState, key: Key
        ) -> tuple[OffPolicyStepState, None]:
            carry = self.step(env, policy, carry, key=key, callback=callback)
            return self.per_step(carry), None

        step_state, _ = filter_scan(
            scan_step, step_state, jr.split(key, self.num_steps)
        )

        return step_state

    @abstractmethod
    def train(
        self,
        policy: PolicyType,
        opt_state: optax.OptState,
        buffer: ReplayBuffer,
        *,
        key: Key,
    ) -> tuple[PolicyType, optax.OptState, dict[str, Scalar]]:
        """
        Trains the policy using data from the replay buffer.

        Args:
            policy: The policy to train.
            opt_state: The current optimizer state.
            buffer: The replay buffer containing experience.
            key: A JAX PRNG key.

        Returns:
            A tuple containing the updated policy, updated optimizer state,
            and a log dictionary with training information.
        """

    def reset(
        self,
        env: AbstractEnvLike,
        policy: PolicyType,
        *,
        key: Key,
        callback: AbstractCallback,
    ) -> OffPolicyState[PolicyType]:
        init_key, starts_key, callback_key = jr.split(key, 3)
        if self.num_envs == 1:
            step_state = OffPolicyStepState.initial(
                self.buffer_size, env, policy, callback, init_key
            )
            step_state = self.collect_learning_starts(
                env, policy, step_state, callback, starts_key
            )
        else:
            step_state = jax.vmap(
                OffPolicyStepState.initial, in_axes=(None, None, None, None, 0)
            )(
                self.buffer_size // self.num_envs,
                env,
                policy,
                callback,
                jr.split(init_key, self.num_envs),
            )
            step_state = jax.vmap(
                self.collect_learning_starts, in_axes=(None, None, 0, None, 0)
            )(env, policy, step_state, callback, jr.split(starts_key, self.num_envs))

        callback_state = callback.reset(ResetContext(locals()), key=callback_key)

        return OffPolicyState(
            jnp.array(0, dtype=int),
            step_state,
            env,
            policy,
            self.optimizer.init(eqx.filter(policy, eqx.is_inexact_array)),
            callback_state,
        )

    def iteration(
        self,
        state: OffPolicyState[PolicyType],
        *,
        key: Key,
        callback: AbstractCallback,
    ) -> OffPolicyState[PolicyType]:
        rollout_key, train_key, callback_key = jr.split(key, 3)

        if self.num_envs == 1:
            step_state = self.collect_rollout(
                state.env, state.policy, state.step_state, callback, rollout_key
            )
        else:
            step_state = eqx.filter_vmap(
                self.collect_rollout, in_axes=(None, None, 0, None, 0)
            )(
                state.env,
                state.policy,
                state.step_state,
                callback,
                jr.split(rollout_key, self.num_envs),
            )

        policy, opt_state, log = self.train(
            state.policy, state.opt_state, step_state.buffer, key=train_key
        )

        state = state.next(step_state, policy, opt_state)

        state = state.with_callback_states(
            callback.on_iteration(
                IterationContext(
                    state.callback_state,
                    state.step_state.callback_state,
                    state.env,
                    state.policy,
                    state.iteration_count,
                    state.opt_state,
                    log,
                    locals(),
                ),
                key=callback_key,
            )
        )

        return self.per_iteration(state)



from __future__ import annotations

import equinox as eqx

from lerax.env import (
    AbstractEnv,
    AbstractEnvLike,
    AbstractEnvLikeState,
    AbstractEnvState,
)
from lerax.render import AbstractRenderer
from lerax.space import AbstractSpace


class AbstractWrapperState[StateType: AbstractEnvLikeState](AbstractEnvLikeState):
    env_state: eqx.AbstractVar[StateType]

    @property
    def unwrapped(self) -> AbstractEnvState:
        """The state of the wrapped environment"""
        return self.env_state.unwrapped


class AbstractWrapper[
    WrapperStateType: AbstractEnvLikeState,
    WrapperActType,
    WrapperObsType,
    StateType: AbstractEnvLikeState,
    ActType,
    ObsType,
](AbstractEnvLike[WrapperStateType, WrapperActType, WrapperObsType]):
    """
    Base class for environment wrappers.

    Attributes:
        name: The name of the environment
        env: The wrapped environment
        unwrapped: The environment without any wrappers
        action_space: The action space of the environment after wrapping
        observation_space: The observation space of the environment after wrapping
    """

    env: eqx.AbstractVar[AbstractEnvLike[StateType, ActType, ObsType]]

    action_space: eqx.AbstractVar[AbstractSpace[ActType]]
    observation_space: eqx.AbstractVar[AbstractSpace[ObsType]]

    @property
    def unwrapped(self) -> AbstractEnv:
        """Return the wrapped environment"""
        return self.env.unwrapped

    @property
    def name(self) -> str:
        """Return the name of the environment"""
        return self.env.name

    def default_renderer(self) -> AbstractRenderer:
        """Return the default renderer for the wrapped environment"""
        return self.unwrapped.default_renderer()

    def render(self, state: WrapperStateType, renderer: AbstractRenderer):
        """Render a frame from a state"""
        self.unwrapped.render(state.unwrapped, renderer)



from __future__ import annotations

from functools import partial
from typing import Callable

import equinox as eqx
from jax import numpy as jnp
from jaxtyping import Array, ArrayLike, Bool, Float, Key

from lerax.env import AbstractEnvLike, AbstractEnvLikeState

from .base_wrapper import AbstractWrapper


class AbstractPureTransformRewardWrapper[
    StateType: AbstractEnvLikeState, ActType, ObsType
](AbstractWrapper[StateType, ActType, ObsType, StateType, ActType, ObsType]):
    """
    Apply a *pure* (stateless) function to every reward emitted by the wrapped
    environment.
    """

    env: eqx.AbstractVar[AbstractEnvLike[StateType, ActType, ObsType]]
    func: eqx.AbstractVar[Callable[[Float[Array, ""]], Float[Array, ""]]]

    def initial(self, *, key: Key) -> StateType:
        return self.env.initial(key=key)

    def transition(self, state: StateType, action: ActType, *, key: Key) -> StateType:
        return self.env.transition(state, action, key=key)

    def observation(self, state: StateType, *, key: Key) -> ObsType:
        return self.env.observation(state, key=key)

    def reward(
        self, state: StateType, action: ActType, next_state: StateType, *, key: Key
    ) -> Float[Array, ""]:
        return self.func(self.env.reward(state, action, next_state, key=key))

    def terminal(self, state: StateType, *, key: Key) -> Bool[Array, ""]:
        return self.env.terminal(state, key=key)

    def truncate(self, state: StateType) -> Bool[Array, ""]:
        return self.env.truncate(state)

    def state_info(self, state: StateType) -> dict:
        return self.env.state_info(state)

    def transition_info(
        self, state: StateType, action: ActType, next_state: StateType
    ) -> dict:
        return self.env.transition_info(state, action, next_state)


class TransformReward[StateType: AbstractEnvLikeState, ActType, ObsType](
    AbstractPureTransformRewardWrapper[StateType, ActType, ObsType]
):
    """
    Apply an arbitrary function to the rewards emitted by the wrapped environment.

    Attributes:
        env: The environment to wrap.
        func: The function to apply to the rewards.

    Args:
        env: The environment to wrap.
        func: The function to apply to the rewards.
    """

    env: AbstractEnvLike[StateType, ActType, ObsType]
    func: Callable[[Float[Array, ""]], Float[Array, ""]]

    def __init__(
        self,
        env: AbstractEnvLike[StateType, ActType, ObsType],
        func: Callable[[Float[Array, ""]], Float[Array, ""]],
    ):
        self.env = env
        self.func = func


class ClipReward[StateType: AbstractEnvLikeState, ActType, ObsType](
    AbstractPureTransformRewardWrapper[StateType, ActType, ObsType]
):
    """
    Cip the rewards emitted by the wrapped environment to a specified range.

    Attributes:
        env: The environment to wrap.
        min: The minimum reward value.
        max: The maximum reward value.

    Args:
        env: The environment to wrap.
        min: The minimum reward value.
        max: The maximum reward value.
    """

    env: AbstractEnvLike[StateType, ActType, ObsType]
    func: Callable[[Float[Array, ""]], Float[Array, ""]]
    min: Float[Array, ""]
    max: Float[Array, ""]

    def __init__(
        self,
        env: AbstractEnvLike[StateType, ActType, ObsType],
        min: Float[ArrayLike, ""] = jnp.asarray(-1.0),
        max: Float[ArrayLike, ""] = jnp.asarray(1.0),
    ):
        self.env = env
        self.min = jnp.asarray(min)
        self.max = jnp.asarray(max)
        self.func = partial(jnp.clip, min=self.min, max=self.max)



from __future__ import annotations

from typing import Callable

import equinox as eqx
from jax import numpy as jnp
from jaxtyping import Array, Float, Key

from lerax.env import AbstractEnvLike, AbstractEnvLikeState
from lerax.space import AbstractSpace, Box

from .base_wrapper import AbstractWrapper
from .utils import rescale_box


class AbstractPureTransformActionWrapper[
    WrapperActType, StateType: AbstractEnvLikeState, ActType, ObsType
](AbstractWrapper[StateType, WrapperActType, ObsType, StateType, ActType, ObsType]):
    """
    Base class for wrappers that apply a pure function to the action before passing it to
    the environment.

    Attributes:
        env: The environment to wrap.
        func: The function to apply to the action.
        action_space: The action space of the wrapper.
    """

    env: eqx.AbstractVar[AbstractEnvLike[StateType, ActType, ObsType]]
    func: eqx.AbstractVar[Callable[[WrapperActType], ActType]]
    action_space: eqx.AbstractVar[AbstractSpace[WrapperActType]]

    @property
    def observation_space(self) -> AbstractSpace[ObsType]:
        return self.env.observation_space

    def initial(self, *, key: Key) -> StateType:
        return self.env.initial(key=key)

    def transition(
        self, state: StateType, action: WrapperActType, *, key: Key
    ) -> StateType:
        return self.env.transition(state, self.func(action), key=key)

    def observation(self, state: StateType, *, key: Key) -> ObsType:
        return self.env.observation(state, key=key)

    def reward(
        self,
        state: StateType,
        action: WrapperActType,
        next_state: StateType,
        *,
        key: Key,
    ) -> Float[Array, ""]:
        return self.env.reward(state, self.func(action), next_state, key=key)

    def terminal(self, state: StateType, *, key: Key) -> jnp.ndarray:
        return self.env.terminal(state, key=key)

    def truncate(self, state: StateType) -> jnp.ndarray:
        return self.env.truncate(state)

    def state_info(self, state: StateType) -> dict:
        return self.env.state_info(state)

    def transition_info(
        self, state: StateType, action: WrapperActType, next_state: StateType
    ) -> dict:
        return self.env.transition_info(state, self.func(action), next_state)


class TransformAction[
    WrapperActType, StateType: AbstractEnvLikeState, ActType, ObsType
](AbstractPureTransformActionWrapper[WrapperActType, StateType, ActType, ObsType]):
    """
    Apply a function to the action before passing it to the environment.

    Attributes:
        env: The environment to wrap.
        func: The function to apply to the action.
        action_space: The action space of the wrapper.

    Args:
        env: The environment to wrap.
        func: The function to apply to the action.
        action_space: The action space of the wrapper.
    """

    env: AbstractEnvLike[StateType, ActType, ObsType]
    func: Callable[[WrapperActType], ActType]
    action_space: AbstractSpace[WrapperActType]

    def __init__(
        self,
        env: AbstractEnvLike[StateType, ActType, ObsType],
        func: Callable[[WrapperActType], ActType],
        action_space: AbstractSpace[WrapperActType],
    ):
        self.env = env
        self.func = func
        self.action_space = action_space


class ClipAction[StateType: AbstractEnvLikeState, ObsType](
    AbstractPureTransformActionWrapper[
        Float[Array, " ..."], StateType, Float[Array, " ..."], ObsType
    ]
):
    """
    Clips every action to the environment's action space.

    Note:
        Only compatible with `Box` action spaces.

    Attributes:
        env: The environment to wrap.
        action_space: The action space of the wrapper.

    Args:
        env: The environment to wrap.

    Raises:
        ValueError: If the environment's action space is not a `Box`.
    """

    env: AbstractEnvLike[StateType, Float[Array, " ..."], ObsType]
    func: Callable[[Float[Array, " ..."]], Float[Array, " ..."]]
    action_space: Box

    def __init__(self, env: AbstractEnvLike[StateType, Float[Array, " ..."], ObsType]):
        if not isinstance(env.action_space, Box):
            raise ValueError(
                "ClipAction only supports `Box` action spaces "
                f"not {type(env.action_space)}"
            )

        def clip(action: Float[Array, " ..."]) -> Float[Array, " ..."]:
            assert isinstance(env.action_space, Box)
            return jnp.clip(action, env.action_space.low, env.action_space.high)

        action_space = Box(-jnp.inf, jnp.inf, shape=env.action_space.shape)

        self.env = env
        self.func = clip
        self.action_space = action_space


class RescaleAction[StateType: AbstractEnvLikeState, ObsType](
    AbstractPureTransformActionWrapper[
        Float[Array, " ..."], StateType, Float[Array, " ..."], ObsType
    ]
):
    """
    Affine rescaling of a box action to a different range.

    Note:
        Only compatible with `Box` action spaces.

    Attributes:
        env: The environment to wrap.
        action_space: The action space of the wrapper.

    Args:
        env: The environment to wrap.

    Raises:
        ValueError: If the environment's action space is not a `Box`.
    """

    env: AbstractEnvLike[StateType, Float[Array, " ..."], ObsType]
    func: Callable[[Float[Array, " ..."]], Float[Array, " ..."]]
    action_space: Box

    def __init__(
        self,
        env: AbstractEnvLike[StateType, Float[Array, " ..."], ObsType],
        min: Float[Array, " ..."] = jnp.array(-1.0),
        max: Float[Array, " ..."] = jnp.array(1.0),
    ):
        if not isinstance(env.action_space, Box):
            raise ValueError(
                "RescaleAction only supports `Box` action spaces"
                f" not {type(env.action_space)}"
            )

        action_space, _, rescale = rescale_box(env.action_space, min, max)

        self.env = env
        self.func = rescale
        self.action_space = action_space



from __future__ import annotations

from typing import ClassVar

import diffrax
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, ArrayLike, Bool, Float, Int, Key

from lerax.render import (
    Abstract2DRenderer,
    AbstractRenderer,
    Color,
    PygameRenderer,
    Transform,
)
from lerax.space import Box, Discrete

from .base_classic_control import (
    AbstractClassicControlEnv,
    AbstractClassicControlEnvState,
)


class MountainCarState(AbstractClassicControlEnvState):
    y: Float[Array, "2"]
    t: Float[Array, ""]


class MountainCar(
    AbstractClassicControlEnv[MountainCarState, Int[Array, ""], Float[Array, "2"]]
):
    """
    Mountain Car environment matching the [Gymnasium MountainCar environment](https://gymnasium.farama.org/environments/classic_control/mountain_car/).

    Note:
        To achieve identical dynamics to Gymnasium set `solver=diffrax.Euler()`.

    ## Action Space

    The action space is discrete with 3 actions:

    - 0: Push left
    - 1: No push
    - 2: Push right

    The action applies a fixed magnitude force to the car in the specified direction for the duration of the time step.

    ## Observation Space

    The observation space is a 2-dimensional continuous space representing the position and velocity of the car:

    | Index | Observation      | Min Value | Max Value |
    |-------|------------------|-----------|-----------|
    | 0     | Car Position     | -1.2      | 0.6       |
    | 1     | Car Velocity     | -0.07     | 0.07      |

    These values reflect the physical limits of the environment.
    These limits can be modified via the `min_position`, `max_position`, and `max_speed` parameters.

    ## Reward

    The reward is -1 for every step taken until the goal position is reached.

    ## Termination

    The episode terminates when the car reaches the goal position of 0.5 with a velocity of at least 0.0.

    Args:
        min_position: Minimum position of the car.
        max_position: Maximum position of the car.
        max_speed: Maximum speed of the car.
        goal_position: Position at which the episode terminates.
        goal_velocity: Minimum velocity at the goal position to terminate the episode.
        force: Magnitude of the force applied when pushing the car.
        gravity: Gravity constant affecting the car's movement.
        dt: Time step for the environment dynamics.
        solver: Diffrax solver to use for integrating the dynamics.
        stepsize_controller: Step size controller for adaptive solvers.
    """

    name: ClassVar[str] = "MountainCar"

    action_space: Discrete
    observation_space: Box

    min_position: Float[Array, ""]
    max_position: Float[Array, ""]
    max_speed: Float[Array, ""]
    goal_position: Float[Array, ""]
    goal_velocity: Float[Array, ""]

    force: Float[Array, ""]
    gravity: Float[Array, ""]
    low: Float[Array, "2"]
    high: Float[Array, "2"]

    dt: Float[Array, ""]
    solver: diffrax.AbstractSolver
    dt0: Float[Array, ""] | None
    stepsize_controller: diffrax.AbstractStepSizeController

    def __init__(
        self,
        *,
        min_position: Float[ArrayLike, ""] = -1.2,
        max_position: Float[ArrayLike, ""] = 0.6,
        max_speed: Float[ArrayLike, ""] = 0.07,
        goal_position: Float[ArrayLike, ""] = 0.5,
        goal_velocity: Float[ArrayLike, ""] = 0.0,
        force: Float[ArrayLike, ""] = 0.001,
        gravity: Float[ArrayLike, ""] = 0.0025,
        dt: Float[ArrayLike, ""] = 1.0,
        solver: diffrax.AbstractSolver | None = None,
        stepsize_controller: diffrax.AbstractStepSizeController | None = None,
    ):
        self.min_position = jnp.array(min_position)
        self.max_position = jnp.array(max_position)
        self.max_speed = jnp.array(max_speed)
        self.goal_position = jnp.array(goal_position)
        self.goal_velocity = jnp.array(goal_velocity)

        self.force = jnp.array(force)
        self.gravity = jnp.array(gravity)

        self.low = jnp.array([self.min_position, -self.max_speed])
        self.high = jnp.array([self.max_position, self.max_speed])

        self.action_space = Discrete(3)
        self.observation_space = Box(self.low, self.high)

        self.dt = jnp.array(dt)
        self.solver = solver or diffrax.Tsit5()
        is_adaptive = isinstance(self.solver, diffrax.AbstractAdaptiveSolver)
        self.dt0 = None if is_adaptive else self.dt
        if stepsize_controller is None:
            self.stepsize_controller = (
                diffrax.PIDController(rtol=1e-5, atol=1e-5)
                if is_adaptive
                else diffrax.ConstantStepSize()
            )
        else:
            self.stepsize_controller = stepsize_controller

    def initial(self, *, key: Key) -> MountainCarState:
        return MountainCarState(
            y=jnp.asarray([jr.uniform(key, minval=-0.6, maxval=-0.4), 0.0]),
            t=jnp.array(0.0),
        )

    def dynamics(
        self, t: Float[Array, ""], y: Float[Array, "2"], action: Int[Array, ""]
    ) -> Float[Array, "2"]:
        x, x_d = y
        u = (action - 1) * self.force
        x_dd = u - self.gravity * jnp.cos(3.0 * x)
        return jnp.array([x_d, x_dd])

    def clip(self, y: Float[Array, "2"]) -> Float[Array, "2"]:
        x, v = y
        v = jnp.clip(v, -self.max_speed, self.max_speed)
        x = jnp.clip(x, self.min_position, self.max_position)
        v = v * ((x != self.min_position) | (v > 0.0))
        return jnp.array([x, v])

    def observation(self, state: MountainCarState, *, key: Key) -> Float[Array, "2"]:
        return state.y

    def reward(
        self,
        state: MountainCarState,
        action: Int[Array, ""],
        next_state: MountainCarState,
        *,
        key: Key,
    ) -> Float[Array, ""]:
        return jnp.array(-1.0)

    def terminal(self, state: MountainCarState, *, key: Key) -> Bool[Array, ""]:
        x, v = state.y
        return (x >= self.goal_position) & (v >= self.goal_velocity)

    def render(self, state: MountainCarState, renderer: AbstractRenderer):
        if not isinstance(renderer, Abstract2DRenderer):
            raise TypeError("MountainCar environment requires a 2D renderer.")

        x = state.y[0]

        renderer.clear()

        # Track
        xs = jnp.linspace(self.min_position - 0.5, self.max_position + 0.1, 64)
        ys = jnp.sin(3 * xs) * 0.45
        track_points = jnp.stack([xs, ys], axis=1)
        track_color = Color(0.0, 0.0, 0.0)
        renderer.draw_polyline(track_points, color=track_color)

        # Flag
        flag_h = 0.2
        flag_start = jnp.array(
            [self.goal_position, jnp.sin(3 * self.goal_position) * 0.45]
        )
        flag_end = flag_start + jnp.array([0.0, flag_h])

        flag_pole_color = Color(0.0, 0.0, 0.0)
        flag_color = Color(0.86, 0.24, 0.24)
        renderer.draw_line(flag_start, flag_end, color=flag_pole_color, width=0.005)
        flag_points = jnp.array(
            [
                flag_end,
                flag_end + jnp.array([0.1, -0.03]),
                flag_end + jnp.array([0.0, -0.06]),
            ]
        )
        renderer.draw_polygon(flag_points, color=flag_color)

        # Car
        car_h, car_w, wheel_r, clearance = 0.04, 0.1, 0.02, 0.025
        angle = jnp.arctan2(jnp.cos(3 * x) * 0.45 * 3, 1.0)
        rot = jnp.array(
            [[jnp.cos(angle), -jnp.sin(angle)], [jnp.sin(angle), jnp.cos(angle)]]
        )
        ## Body
        car_col = Color(0.0, 0.0, 0.0)
        clearance_vec = rot @ jnp.array([0.0, clearance + car_h / 2])
        car_center = jnp.array([x, jnp.sin(3 * x) * 0.45]) + clearance_vec

        car_corners = jnp.array(
            [
                [-car_w / 2, -car_h / 2],
                [car_w / 2, -car_h / 2],
                [car_w / 2, car_h / 2],
                [-car_w / 2, car_h / 2],
            ]
        )
        car_corners = (rot @ car_corners.T).T + car_center
        renderer.draw_polygon(car_corners, color=car_col)
        ## Wheels
        wheel_col = Color(0.3, 0.3, 0.3)
        wheel_clearance = rot @ jnp.array([0.0, wheel_r])
        wheel_centers = jnp.array(
            [
                [-car_w / 4, 0.0],
                [car_w / 4, 0.0],
            ]
        )
        wheel_centers = (
            (rot @ wheel_centers.T).T
            + jnp.array([x, jnp.sin(3 * x) * 0.45])
            + wheel_clearance
        )
        renderer.draw_circle(wheel_centers[0], radius=wheel_r, color=wheel_col)
        renderer.draw_circle(wheel_centers[1], radius=wheel_r, color=wheel_col)

        renderer.draw()

    def default_renderer(self) -> Abstract2DRenderer:
        width, height = 800, 450
        transform = Transform(
            scale=350.0,
            offset=jnp.array([width * 0.7, height * 0.4]),
            width=width,
            height=height,
            y_up=True,
        )
        return PygameRenderer(
            width=width,
            height=height,
            background_color=Color(1.0, 1.0, 1.0),
            transform=transform,
        )



from __future__ import annotations

from abc import abstractmethod
from typing import Concatenate

import equinox as eqx


class AbstractModelState(eqx.Module):
    pass


class AbstractModel[**InType, OutType](eqx.Module):
    """Base class for models that take inputs and produce outputs."""

    @abstractmethod
    def __call__(self, *args: InType.args, **kwargs: InType.kwargs) -> OutType:
        """Return an output given an input."""


class AbstractStatefulModel[StateType: AbstractModelState, **InType, *OutType](
    AbstractModel[Concatenate[StateType, InType], tuple[StateType, *OutType]],
):
    """Base class for models with state."""

    @abstractmethod
    def reset(self) -> StateType:
        """Reset the model to its initial state."""



from __future__ import annotations

from jax import numpy as jnp
from jaxtyping import Array, ArrayLike, Bool, Float, Int, Key

from lerax.env import AbstractEnvLike, AbstractEnvLikeState
from lerax.space import AbstractSpace

from .base_wrapper import (
    AbstractWrapper,
    AbstractWrapperState,
)


class Identity[StateType: AbstractEnvLikeState, ActType, ObsType](
    AbstractWrapper[StateType, ActType, ObsType, StateType, ActType, ObsType]
):
    """
    An wrapper that does nothing.

    Attributes:
        env: The environment to wrap.

    Args:
        env: The environment to wrap.
    """

    env: AbstractEnvLike[StateType, ActType, ObsType]

    def __init__(self, env: AbstractEnvLike[StateType, ActType, ObsType]):
        self.env = env

    @property
    def action_space(self) -> AbstractSpace[ActType]:
        return self.env.action_space

    @property
    def observation_space(self) -> AbstractSpace[ObsType]:
        return self.env.observation_space

    def initial(self, *, key: Key) -> StateType:
        return self.env.initial(key=key)

    def transition(self, state: StateType, action: ActType, *, key: Key) -> StateType:
        return self.env.transition(state, action, key=key)

    def observation(self, state: StateType, *, key: Key) -> ObsType:
        return self.env.observation(state, key=key)

    def reward(
        self, state: StateType, action: ActType, next_state: StateType, *, key: Key
    ) -> Float[Array, ""]:
        return self.env.reward(state, action, next_state, key=key)

    def terminal(self, state: StateType, *, key: Key) -> Bool[Array, ""]:
        return self.env.terminal(state, key=key)

    def truncate(self, state: StateType) -> Bool[Array, ""]:
        return self.env.truncate(state)

    def state_info(self, state: StateType) -> dict:
        return self.env.state_info(state)

    def transition_info(
        self, state: StateType, action: ActType, next_state: StateType
    ) -> dict:
        return self.env.transition_info(state, action, next_state)


class TimeLimitState[StateType: AbstractEnvLikeState](AbstractWrapperState):
    env_state: StateType
    step_count: Int[Array, ""]

    def __init__(self, step_count: Int[ArrayLike, ""], env_state: StateType):
        self.step_count = jnp.array(step_count, dtype=int)
        self.env_state = env_state


class TimeLimit[StateType: AbstractEnvLikeState, ActType, ObsType](
    AbstractWrapper[
        TimeLimitState[StateType], ActType, ObsType, StateType, ActType, ObsType
    ]
):
    """
    Time limit wrapper that truncates episodes after a fixed number of steps.

    Attributes:
        env: The environment to wrap.
        max_episode_steps: The maximum number of steps per episode.

    Args:
        env: The environment to wrap.
        max_episode_steps: The maximum number of steps per episode.
    """

    env: AbstractEnvLike[StateType, ActType, ObsType]
    max_episode_steps: Int[Array, ""]

    def __init__(
        self, env: AbstractEnvLike[StateType, ActType, ObsType], max_episode_steps: int
    ):
        self.env = env
        self.max_episode_steps = jnp.array(max_episode_steps, dtype=int)

    @property
    def action_space(self) -> AbstractSpace[ActType]:
        return self.env.action_space

    @property
    def observation_space(self) -> AbstractSpace[ObsType]:
        return self.env.observation_space

    def initial(self, *, key: Key) -> TimeLimitState[StateType]:
        env_state = self.env.initial(key=key)
        return TimeLimitState(step_count=0, env_state=env_state)

    def transition(
        self, state: TimeLimitState[StateType], action: ActType, *, key: Key
    ) -> TimeLimitState[StateType]:
        env_next_state = self.env.transition(state.env_state, action, key=key)
        return TimeLimitState(step_count=state.step_count + 1, env_state=env_next_state)

    def observation(self, state: TimeLimitState[StateType], *, key: Key) -> ObsType:
        return self.env.observation(state.env_state, key=key)

    def reward(
        self,
        state: TimeLimitState[StateType],
        action: ActType,
        next_state: TimeLimitState[StateType],
        *,
        key: Key,
    ) -> Float[Array, ""]:
        return self.env.reward(state.env_state, action, next_state.env_state, key=key)

    def terminal(
        self, state: TimeLimitState[StateType], *, key: Key
    ) -> Bool[Array, ""]:
        return self.env.terminal(state.env_state, key=key)

    def truncate(self, state: TimeLimitState[StateType]) -> Bool[Array, ""]:
        env_truncate = self.env.truncate(state.env_state)
        return env_truncate | (state.step_count >= self.max_episode_steps)

    def state_info(self, state: TimeLimitState[StateType]) -> dict:
        return self.env.state_info(state.env_state)

    def transition_info(
        self,
        state: TimeLimitState[StateType],
        action: ActType,
        next_state: TimeLimitState[StateType],
    ) -> dict:
        return self.env.transition_info(state.env_state, action, next_state.env_state)



from __future__ import annotations

from typing import Any

import numpy as np
from jax import numpy as jnp
from jaxtyping import ArrayLike, Float

from .base_renderer import Abstract2DRenderer, Color, Transform

imageio: Any | None = None


def _load_imageio():
    global imageio
    if imageio is not None:
        return imageio
    try:
        import imageio.v2 as _imageio
    except ImportError as exc:
        raise ImportError(
            "VideoRenderer requires the optional rendering dependencies. "
            "Install them with: pip install lerax[render]"
        ) from exc
    imageio = _imageio
    return imageio


class VideoRenderer(Abstract2DRenderer):
    """
    Renderer wrapper that records frames to a video file.

    It forwards all drawing operations to the underlying renderer and captures
    the framebuffer on each call to `draw()`. When `close()` is called, the
    accumulated frames are written to `output_path` as a video.

    Note:
        This renderer is not JIT-safe and is intended for use in Python loops
        (e.g. via `env.render_states` / `env.render_stacked`).

    Attributes:
        transform: Transform from world space to pixel space.
        inner: The underlying renderer to which drawing operations are forwarded.
        output_path: Path to the output video file.
        fps: Frames per second for the output video.
        frames: List of captured frames.

    Args:
        inner: The underlying renderer to which drawing operations are forwarded.
        output_path: Path to the output video file.
        fps: Frames per second for the output video.
    """

    transform: Transform

    inner: Abstract2DRenderer
    output_path: str
    fps: float

    frames: list

    def __init__(
        self,
        inner: Abstract2DRenderer,
        output_path: str,
        fps: float = 60.0,
    ) -> None:
        self.inner = inner
        self.transform = inner.transform
        self.output_path = output_path
        self.fps = fps
        self.frames = []

    def open(self):
        self.inner.open()

    def close(self):
        if self.frames:
            writer = _load_imageio().get_writer(self.output_path, fps=self.fps)
            try:
                for frame in self.frames:
                    arr = jnp.asarray(frame)
                    arr = jnp.clip(arr, 0, 255).astype(jnp.uint8)
                    writer.append_data(np.array(arr))
            finally:
                writer.close()
        self.inner.close()

    def draw(self):
        self.inner.draw()
        frame = self.inner.as_array()
        self.frames.append(frame)

    def clear(self):
        self.inner.clear()

    def draw_circle(
        self,
        center: Float[ArrayLike, "2"],
        radius: Float[ArrayLike, ""],
        color: Color,
    ):
        self.inner.draw_circle(center, radius, color)

    def draw_line(
        self,
        start: Float[ArrayLike, "2"],
        end: Float[ArrayLike, "2"],
        color: Color,
        width: Float[ArrayLike, ""] = 1,
    ):
        self.inner.draw_line(start, end, color, width)

    def draw_rect(
        self,
        center: Float[ArrayLike, "2"],
        w: Float[ArrayLike, ""],
        h: Float[ArrayLike, ""],
        color: Color,
    ):
        self.inner.draw_rect(center, w, h, color)

    def draw_polygon(
        self,
        points: Float[ArrayLike, "num 2"],
        color: Color,
    ):
        self.inner.draw_polygon(points, color)

    def draw_text(
        self,
        center: Float[ArrayLike, "2"],
        text: str,
        color: Color,
        size: Float[ArrayLike, ""] = 12,
    ):
        self.inner.draw_text(center, text, color, size)

    def draw_polyline(
        self,
        points: Float[ArrayLike, "num 2"],
        color: Color,
    ):
        self.inner.draw_polyline(points, color)

    def draw_ellipse(
        self,
        center: Float[ArrayLike, "2"],
        w: Float[ArrayLike, ""],
        h: Float[ArrayLike, ""],
        color: Color,
    ):
        self.inner.draw_ellipse(center, w, h, color)

    def as_array(self) -> Float[ArrayLike, "H W 3"]:
        return self.inner.as_array()



from .base_algorithm import AbstractAlgorithm, AbstractAlgorithmState, AbstractStepState
from .dqn import DQN
from .off_policy import AbstractOffPolicyAlgorithm, OffPolicyState, OffPolicyStepState
from .on_policy import AbstractOnPolicyAlgorithm, OnPolicyState, OnPolicyStepState
from .ppo import PPO

__all__ = [
    "AbstractAlgorithm",
    "AbstractAlgorithmState",
    "AbstractStepState",
    "AbstractOffPolicyAlgorithm",
    "OffPolicyState",
    "OffPolicyStepState",
    "AbstractOnPolicyAlgorithm",
    "OnPolicyState",
    "OnPolicyStepState",
    "PPO",
    "DQN",
]



from __future__ import annotations

import equinox as eqx
import jax
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, ArrayLike, Bool, Float, Integer, Key, PyTree

from lerax.policy import AbstractPolicyState
from lerax.space import AbstractSpace

from .base_buffer import AbstractBuffer


class ReplayBuffer[StateType: AbstractPolicyState, ActType, ObsType](AbstractBuffer):
    size: int = eqx.field(static=True)
    position: Integer[Array, ""]

    observations: PyTree[ObsType]
    next_observations: PyTree[ObsType]
    actions: PyTree[ActType]
    rewards: Float[Array, " capacity"]
    dones: Bool[Array, " capacity"]
    timeouts: Bool[Array, " capacity"]
    states: StateType
    next_states: StateType

    def __init__(
        self,
        size: int,
        observation_space: AbstractSpace[ObsType],
        action_space: AbstractSpace[ActType],
        state: StateType,
    ):
        self.size = size
        self.position = jnp.array(0, dtype=int)

        def init_leaf(example):
            arr = jnp.asarray(example)
            return jnp.broadcast_to(arr, (self.size,) + arr.shape)

        self.observations = jax.tree.map(init_leaf, observation_space.canonical())
        self.next_observations = jax.tree.map(init_leaf, observation_space.canonical())
        self.actions = jax.tree.map(init_leaf, action_space.canonical())

        self.rewards = jnp.zeros((self.size,), dtype=float)
        self.dones = jnp.zeros((self.size,), dtype=bool)
        self.timeouts = jnp.zeros((self.size,), dtype=bool)

        self.states = jax.tree.map(init_leaf, state)
        self.next_states = jax.tree.map(init_leaf, state)

    @property
    def shape(self) -> tuple[int, ...]:
        return self.rewards.shape

    @property
    def current_size(self) -> Integer[Array, ""]:
        return jnp.minimum(self.position, self.size)

    def add(
        self,
        observation: ObsType,
        next_observation: ObsType,
        action: ActType,
        reward: Float[ArrayLike, ""],
        done: Bool[ArrayLike, ""],
        timeout: Bool[ArrayLike, ""],
        state: StateType,
        next_state: StateType,
    ) -> ReplayBuffer[StateType, ActType, ObsType]:
        reward = jnp.asarray(reward, dtype=float)
        done = jnp.asarray(done, dtype=bool)
        timeout = jnp.asarray(timeout, dtype=bool)

        idx = self.position % self.size

        def set_at_idx(leaf, new_value):
            return leaf.at[idx].set(new_value)

        observations = jax.tree.map(set_at_idx, self.observations, observation)
        next_observations = jax.tree.map(
            set_at_idx, self.next_observations, next_observation
        )
        actions = jax.tree.map(set_at_idx, self.actions, action)
        rewards = self.rewards.at[idx].set(reward)
        dones = self.dones.at[idx].set(done)
        timeouts = self.timeouts.at[idx].set(timeout)
        states = jax.tree.map(set_at_idx, self.states, state)
        next_states = jax.tree.map(set_at_idx, self.next_states, next_state)

        new_position = self.position + 1

        return eqx.tree_at(
            lambda rb: (
                rb.position,
                rb.observations,
                rb.next_observations,
                rb.actions,
                rb.rewards,
                rb.dones,
                rb.timeouts,
                rb.states,
                rb.next_states,
            ),
            self,
            (
                new_position,
                observations,
                next_observations,
                actions,
                rewards,
                dones,
                timeouts,
                states,
                next_states,
            ),
        )

    def batches(
        self,
        batch_size: int,
        *,
        key: Key | None = None,
        batch_axes: tuple[int, ...] | int | None = None,
    ) -> ReplayBuffer[StateType, ActType, ObsType]:
        _ = eqx.error_if(
            self.current_size,
            self.current_size < self.size,
            "ReplayBuffer.batches assumes the buffer is full.",
        )

        flat_self = self.flatten_axes(batch_axes)

        total = flat_self.rewards.shape[0]
        indices = jnp.arange(total) if key is None else jr.permutation(key, total)

        if total % batch_size != 0:
            total_trim = total - (total % batch_size)
            indices = indices[:total_trim]

        indices = indices.reshape(-1, batch_size)

        def take_batch(x):
            if not isinstance(x, jnp.ndarray) or x.ndim == 0:
                return x
            return jnp.take(x, indices, axis=0)

        return jax.tree.map(take_batch, flat_self)

    def sample(
        self,
        batch_size: int,
        *,
        key: Key,
    ) -> ReplayBuffer[StateType, ActType, ObsType]:
        flat_self = self.flatten_axes(None)
        total = flat_self.rewards.shape[0]

        current_size = self.current_size

        if current_size.ndim == 0:
            valid_mask = jnp.arange(self.size) < current_size
        else:
            valid_mask = (jnp.arange(self.size) < current_size[..., None]).reshape(-1)

        probs = valid_mask.astype(float) / jnp.sum(valid_mask)
        batch_indices = jr.choice(
            key,
            total,
            shape=(batch_size,),
            replace=False,
            p=probs,
        )

        def take_sample(x):
            if not isinstance(x, jnp.ndarray) or x.ndim == 0:
                return x
            return jnp.take(x, batch_indices, axis=0)

        batch = jax.tree.map(take_sample, flat_self)

        return batch



from .base_q import (
    AbstractQPolicy,
    AbstractStatefulQPolicy,
    AbstractStatelessQPolicy,
    QStatefulWrapper,
)
from .mlp import MLPQPolicy

__all__ = [
    "AbstractQPolicy",
    "AbstractStatefulQPolicy",
    "AbstractStatelessQPolicy",
    "MLPQPolicy",
    "QStatefulWrapper",
]



from .acrobot import Acrobot
from .cartpole import CartPole
from .continuous_mountain_car import ContinuousMountainCar
from .mountain_car import MountainCar
from .pendulum import Pendulum

__all__ = ["Acrobot", "CartPole", "MountainCar", "ContinuousMountainCar", "Pendulum"]



from .actor_critic import (
    AbstractActorCriticPolicy,
    AbstractStatefulActorCriticPolicy,
    AbstractStatelessActorCriticPolicy,
    ActorCriticStatefulWrapper,
    MLPActorCriticPolicy,
    NCDEActorCriticPolicy,
)
from .base_policy import (
    AbstractPolicy,
    AbstractPolicyState,
    AbstractStatefulPolicy,
    AbstractStatefulWrapper,
    AbstractStatelessPolicy,
)
from .dqn import (
    AbstractDQNPolicy,
    AbstractStatefulDQNPolicy,
    AbstractStatelessDQNPolicy,
    DQNStatefulWrapper,
    MLPDQNPolicy,
)
from .q import (
    AbstractQPolicy,
    AbstractStatefulQPolicy,
    AbstractStatelessQPolicy,
    MLPQPolicy,
    QStatefulWrapper,
)

__all__ = [
    "AbstractPolicy",
    "AbstractPolicyState",
    "AbstractStatefulPolicy",
    "AbstractStatelessPolicy",
    "AbstractStatefulWrapper",
    "AbstractActorCriticPolicy",
    "AbstractStatefulActorCriticPolicy",
    "AbstractStatelessActorCriticPolicy",
    "ActorCriticStatefulWrapper",
    "MLPActorCriticPolicy",
    "NCDEActorCriticPolicy",
    "AbstractQPolicy",
    "AbstractStatelessQPolicy",
    "AbstractStatefulQPolicy",
    "MLPQPolicy",
    "QStatefulWrapper",
    "AbstractDQNPolicy",
    "AbstractStatelessDQNPolicy",
    "AbstractStatefulDQNPolicy",
    "DQNStatefulWrapper",
    "MLPDQNPolicy",
]



"""
Lerax Distributions

Wrapper around distreqx.distributions to allow for easier imports, extended typing, and
future expansion.
"""

from .base_distribution import (
    AbstractDistreqxWrapper,
    AbstractDistribution,
    AbstractMaskableDistribution,
    AbstractTransformedDistribution,
)
from .bernoulli import Bernoulli
from .categorical import Categorical
from .multi_categorical import MultiCategorical
from .multivariate_normal import MultivariateNormalDiag
from .normal import Normal
from .squashed_multivariate_normal import SquashedMultivariateNormalDiag
from .squashed_normal import SquashedNormal

__all__ = [
    "AbstractDistribution",
    "AbstractDistreqxWrapper",
    "AbstractMaskableDistribution",
    "AbstractTransformedDistribution",
    "Bernoulli",
    "Categorical",
    "Normal",
    "MultivariateNormalDiag",
    "SquashedMultivariateNormalDiag",
    "SquashedNormal",
    "MultiCategorical",
]



from .base_wrapper import AbstractWrapper, AbstractWrapperState
from .misc import Identity, TimeLimit
from .transform_action import ClipAction, RescaleAction, TransformAction
from .transform_observation import (
    ClipObservation,
    FlattenObservation,
    RescaleObservation,
    TransformObservation,
)
from .transform_reward import ClipReward, TransformReward

__all__ = [
    "AbstractWrapper",
    "AbstractWrapperState",
    "Identity",
    "TimeLimit",
    "TransformAction",
    "ClipAction",
    "RescaleAction",
    "ClipObservation",
    "FlattenObservation",
    "RescaleObservation",
    "TransformObservation",
    "ClipReward",
    "TransformReward",
]



from __future__ import annotations

import warnings
from abc import abstractmethod
from typing import Self

import equinox as eqx
from distreqx import bijectors, distributions
from jaxtyping import Array, Bool, Float, Key


class AbstractDistribution[SampleType](eqx.Module):
    """
    Base class for all distributions in Lerax.
    """

    @abstractmethod
    def log_prob(self, value: SampleType) -> Float[Array, ""]:
        """Compute the log probability of a sample."""

    @abstractmethod
    def prob(self, value: SampleType) -> Float[Array, ""]:
        """Compute the probability of a sample."""

    @abstractmethod
    def sample(self, key: Key) -> SampleType:
        """Return a sample from the distribution."""

    @abstractmethod
    def entropy(self) -> Float[Array, ""]:
        """Compute the entropy of the distribution."""

    @abstractmethod
    def mean(self) -> SampleType:
        """Compute the mean of the distribution."""

    @abstractmethod
    def mode(self) -> SampleType:
        """Compute the mode of the distribution."""

    @abstractmethod
    def sample_and_log_prob(self, key: Key) -> tuple[SampleType, Float[Array, ""]]:
        """Return a sample and its log probability."""


class AbstractDistreqxWrapper[SampleType](AbstractDistribution):
    """
    Base class for distributions that wrap distreqx distributions in Lerax.

    Attributes:
        distribution: The underlying distreqx distribution.
    """

    distribution: eqx.AbstractVar[distributions.AbstractDistribution]

    def log_prob(self, value: SampleType) -> Float[Array, ""]:
        return self.distribution.log_prob(value)

    def prob(self, value: SampleType) -> Float[Array, ""]:
        return self.distribution.prob(value)

    def sample(self, key: Key) -> SampleType:
        return self.distribution.sample(key)

    def entropy(self) -> Float[Array, ""]:
        return self.distribution.entropy()

    def mean(self) -> SampleType:
        return self.distribution.mean()

    def mode(self) -> SampleType:
        return self.distribution.mode()

    def sample_and_log_prob(self, key: Key) -> tuple[SampleType, Float[Array, ""]]:
        return self.distribution.sample_and_log_prob(key)


class AbstractMaskableDistribution[SampleType](AbstractDistribution[SampleType]):
    """
    Base class for all maskable distributions in Lerax.

    Maskable distributions allow masking of elements in the distribution.

    Attributes:
        distribution: The underlying distreqx distribution.
    """

    @abstractmethod
    def mask(self, mask: Bool[Array, "..."]) -> Self:
        """
        Return a masked version of the distribution.

        A masked distribution only considers the elements where the mask is True.

        Args:
            mask: A boolean array indicating which elements to keep.

        Returns:
            A new masked distribution.
        """


class AbstractTransformedDistribution[SampleType](AbstractDistreqxWrapper[SampleType]):
    """
    Base class for all transformed distributions in Lerax.

    Transformed distributions apply a bijective transformation to a base distribution.

    Attributes:
        distribution: The underlying distreqx transformed distribution.
        bijector: The bijective transformation applied to the base distribution.
    """

    distribution: eqx.AbstractVar[distributions.AbstractTransformed]

    # This breaks from the abstract/formal pattern but I think it's justified
    def mode(self) -> SampleType:
        try:
            return self.distribution.mode()
        except NotImplementedError:
            # Computing the mode this way is not always correct, but it is a reasonable workaround for the
            # use cases of this library.
            warnings.warn(
                "Mode not implemented for base distribution; using bijector to compute mode."
            )
            return self.distribution._bijector.forward(
                self.distribution._distribution.mode()
            )

    @property
    def bijector(self) -> bijectors.AbstractBijector:
        return self.distribution.bijector



from .base_env import (
    AbstractEnv,
    AbstractEnvLike,
    AbstractEnvLikeState,
    AbstractEnvState,
)
from .classic_control import (
    Acrobot,
    CartPole,
    ContinuousMountainCar,
    MountainCar,
    Pendulum,
)

__all__ = [
    "AbstractEnvLike",
    "AbstractEnv",
    "AbstractEnvLikeState",
    "AbstractEnvState",
    "Acrobot",
    "CartPole",
    "MountainCar",
    "ContinuousMountainCar",
    "Pendulum",
]



from __future__ import annotations

from abc import abstractmethod

import equinox as eqx
from jaxtyping import Key

from lerax.space.base_space import AbstractSpace
from lerax.utils import Serializable


class AbstractPolicyState(eqx.Module):
    """
    Base class for policy internal states.
    """

    pass


class AbstractStatelessPolicy[ActType, ObsType](Serializable):
    """
    Base class for stateless policies.

    Policies map observations to actions.

    Attributes:
        name: The name of the policy.
        action_space: The action space of the policy.
        observation_space: The observation space of the policy.
    """

    name: eqx.AbstractClassVar[str]
    action_space: eqx.AbstractVar[AbstractSpace[ActType]]
    observation_space: eqx.AbstractVar[AbstractSpace[ObsType]]

    @abstractmethod
    def __call__(self, observation: ObsType, *, key: Key | None = None) -> ActType:
        pass

    @abstractmethod
    def into_stateful[SelfType: AbstractStatelessPolicy](
        self: SelfType,
    ) -> AbstractStatefulWrapper[SelfType, ActType, ObsType]:
        pass


class AbstractStatefulPolicy[StateType: AbstractPolicyState, ActType, ObsType](
    Serializable
):
    """
    Base class for stateful policies.

    Policies map observations and internal states to actions and new internal states.

    Attributes:
        name: The name of the policy.
        action_space: The action space of the policy.
        observation_space: The observation space of the policy.
    """

    name: eqx.AbstractClassVar[str]
    action_space: eqx.AbstractVar[AbstractSpace[ActType]]
    observation_space: eqx.AbstractVar[AbstractSpace[ObsType]]

    @abstractmethod
    def __call__(
        self, state: StateType, observation: ObsType, *, key: Key | None = None
    ) -> tuple[StateType, ActType]:
        pass

    @abstractmethod
    def reset(self, *, key: Key) -> StateType:
        pass

    def into_stateful[SelfType: AbstractStatefulPolicy](self: SelfType) -> SelfType:
        return self


class NullStatefulPolicyState(AbstractPolicyState):
    pass


class AbstractStatefulWrapper[PolicyType: AbstractStatelessPolicy, ActType, ObsType](
    AbstractStatefulPolicy[NullStatefulPolicyState, ActType, ObsType]
):
    """
    Wrapper to convert a stateless policy into a stateful policy with a null state.

    Used to enable compatibility between stateless and stateful policy interfaces.

    Attributes:
        name: The name of the policy.
        action_space: The action space of the policy.
        observation_space: The observation space of the policy.
        policy: The underlying stateless policy.
    """

    policy: eqx.AbstractVar[PolicyType]

    @property
    def name(self) -> str:
        return self.policy.name

    @property
    def action_space(self) -> AbstractSpace[ActType]:
        return self.policy.action_space

    @property
    def observation_space(self) -> AbstractSpace[ObsType]:
        return self.policy.observation_space

    def __call__(
        self,
        state: NullStatefulPolicyState,
        observation: ObsType,
        *,
        key: Key | None = None,
    ) -> tuple[NullStatefulPolicyState, ActType]:
        action = self.policy(observation, key=key)
        return state, action

    def reset(self, *, key: Key) -> NullStatefulPolicyState:
        return NullStatefulPolicyState()

    def into_stateless(self) -> PolicyType:
        return self.policy


type AbstractPolicy = AbstractStatelessPolicy | AbstractStatefulPolicy



from __future__ import annotations

from abc import abstractmethod

import equinox as eqx
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, ArrayLike, Float, Int, Key

from lerax.distribution import (
    AbstractDistribution,
    Bernoulli,
    Categorical,
    MultiCategorical,
    MultivariateNormalDiag,
    Normal,
)
from lerax.space import AbstractSpace, Box, Discrete, MultiBinary, MultiDiscrete

from .base_model import AbstractModel


class AbstractActionDistribution[ActType](AbstractModel):
    """Layer that produces action distributions given inputs."""

    mapping: eqx.AbstractVar[eqx.nn.Linear]

    @abstractmethod
    def __call__(
        self,
        inputs: Float[Array, " latent_dim"],
    ) -> AbstractDistribution[ActType]:
        """Produce an action distribution given inputs."""


class BoxAction(AbstractActionDistribution[Float[Array, " action_dim"]]):

    scalar: bool
    mapping: eqx.nn.Linear
    log_std: Float[Array, " *action_dim"]

    def __init__(
        self,
        latent_dim: int,
        action_space: Box,
        *,
        key: Key,
        log_std_init: Float[ArrayLike, ""] = jnp.array(0.0),
    ):
        if action_space.shape:
            self.scalar = False
            self.mapping = eqx.nn.Linear(
                latent_dim, int(action_space.flat_size), key=key
            )
            self.log_std = jnp.full(int(action_space.flat_size), log_std_init)
        else:
            self.scalar = True
            self.mapping = eqx.nn.Linear(latent_dim, "scalar", key=key)
            self.log_std = jnp.array(log_std_init)

    def __call__(
        self, inputs: Float[Array, " latent_dim"]
    ) -> Normal | MultivariateNormalDiag:
        if self.scalar:
            return Normal(
                loc=self.mapping(inputs),
                scale=jnp.exp(self.log_std),
            )
        else:
            return MultivariateNormalDiag(
                loc=self.mapping(inputs),
                scale_diag=jnp.exp(self.log_std),
            )


class DiscreteAction(AbstractActionDistribution[Int[Array, ""]]):

    mapping: eqx.nn.Linear

    def __init__(self, latent_dim: int, action_space: Discrete, *, key: Key):
        self.mapping = eqx.nn.Linear(latent_dim, int(action_space.n), key=key)

    def __call__(self, inputs: Float[Array, " latent_dim"]) -> Categorical:
        return Categorical(logits=self.mapping(inputs))


class MultiBinaryAction(AbstractActionDistribution[Int[Array, ""]]):

    mapping: eqx.nn.Linear
    shape: tuple[int, ...]

    def __init__(self, latent_dim: int, action_space: MultiBinary, *, key: Key):
        self.mapping = eqx.nn.Linear(latent_dim, action_space.flat_size, key=key)
        self.shape = action_space.shape

    def __call__(self, inputs: Float[Array, " latent_dim"]) -> Bernoulli:
        return Bernoulli(logits=self.mapping(inputs).reshape(self.shape))


class MultiDiscreteAction(AbstractActionDistribution[Int[Array, ""]]):

    ns: tuple[int, ...]
    mappings: eqx.nn.Linear
    shape: tuple[int, ...]

    def __init__(self, latent_dim: int, action_space: MultiDiscrete, *, key: Key):
        self.ns = action_space.ns
        self.mappings = eqx.nn.Linear(latent_dim, sum(action_space.ns), key=key)
        self.shape = action_space.shape

    def __call__(self, inputs: Float[Array, " latent_dim"]) -> MultiCategorical:
        return MultiCategorical(self.mappings(inputs), action_dims=self.ns)


def make_action_layer(
    latent_dim: int, action_space: AbstractSpace, *, key: Key, log_std_init: float = 0.0
) -> AbstractActionDistribution:
    """Create an action layer based on the action space."""

    if isinstance(action_space, Box):
        return BoxAction(latent_dim, action_space, key=key, log_std_init=log_std_init)
    elif isinstance(action_space, Discrete):
        return DiscreteAction(latent_dim, action_space, key=key)
    elif isinstance(action_space, MultiBinary):
        return MultiBinaryAction(latent_dim, action_space, key=key)
    elif isinstance(action_space, MultiDiscrete):
        return MultiDiscreteAction(latent_dim, action_space, key=key)
    else:
        raise NotImplementedError(f"Action space {type(action_space)} not supported.")


class ActionLayer[ActType](AbstractModel):
    """
    Model that produces action distributions from features.

    An optional MLP processes the features before passing them to the action distribution layer.
    If depth is set to 1, no MLP is used and the features are affine-mapped directly to the action parameters.

    Attributes:
        mlp: Optional MLP to process features before action distribution.
        action_dist: Action distribution layer.

    Args:
        action_space: The action space defining the type of actions.
        latent_dim: Dimension of the input features.
        width_size: Width of the hidden layers in the MLP.
        depth: Depth of the MLP. If set to 1, no MLP is
        log_std_init: Initial log standard deviation for continuous action spaces.
        key: JAX PRNG key for parameter initialization.
    """

    mlp: eqx.nn.MLP | None
    action_dist: AbstractActionDistribution[ActType]

    def __init__(
        self,
        action_space: AbstractSpace,
        latent_dim: int,
        width_size: int,
        depth: int,
        *,
        key: Key,
        log_std_init: float = 0.0,
    ):
        mlp_key, action_key = jr.split(key, 2)

        depth = max(0, depth - 1)
        if depth < 0:
            self.mlp = None
        else:
            self.mlp = eqx.nn.MLP(
                in_size=latent_dim,
                out_size=latent_dim,
                width_size=width_size,
                depth=depth,
                key=mlp_key,
            )
        self.action_dist = make_action_layer(
            latent_dim, action_space, key=action_key, log_std_init=log_std_init
        )

    def __call__(
        self, features: Float[Array, " latent_dim"]
    ) -> AbstractDistribution[ActType]:
        if self.mlp is not None:
            features = self.mlp(features)
        return self.action_dist(features)



from __future__ import annotations

try:
    from flax import struct
    from gymnax.environments import environment as gym
    from gymnax.environments import spaces as gym_spaces
except ModuleNotFoundError as e:
    raise ImportError(
        "Gymnax compatibility requires toptional dependencies. "
        "Install with `pip install lerax[compatibility]` or install "
        "`gymnax` and `flax` manually."
    ) from e
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, ArrayLike, Bool, Float, Int, Key

from lerax.env import AbstractEnv, AbstractEnvState
from lerax.render import AbstractRenderer
from lerax.space import AbstractSpace, Box, Dict, Discrete, Tuple


def gymnax_space_to_lerax_space(space: gym_spaces.Space) -> AbstractSpace:
    """
    Returns a Lerax space corresponding to the given Gymnax space.

    Args:
        space: Gymnax space to convert.

    Returns:
        The corresponding Lerax space.
    """
    if isinstance(space, gym_spaces.Discrete):
        return Discrete(n=space.n)
    elif isinstance(space, gym_spaces.Box):
        return Box(low=space.low, high=space.high, shape=space.shape)
    elif isinstance(space, gym_spaces.Dict):
        return Dict(
            {k: gymnax_space_to_lerax_space(v) for k, v in space.spaces.items()}
        )
    elif isinstance(space, gym_spaces.Tuple):
        return Tuple(tuple(gymnax_space_to_lerax_space(s) for s in space.spaces))
    else:
        raise NotImplementedError(f"Space type {type(space)} not supported")


def lerax_to_gymnax_space(space: AbstractSpace) -> gym_spaces.Space:
    """
    Returns a Gymnax space corresponding to the given Lerax space.

    Args:
        space: Lerax space to convert.

    Returns:
        The corresponding Gymnax space.
    """
    if isinstance(space, Discrete):
        return gym_spaces.Discrete(int(space.n))
    elif isinstance(space, Box):
        return gym_spaces.Box(low=space.low, high=space.high, shape=space.shape)
    elif isinstance(space, Dict):
        return gym_spaces.Dict(
            {k: lerax_to_gymnax_space(v) for k, v in space.spaces.items()}
        )
    elif isinstance(space, Tuple):
        return gym_spaces.Tuple(tuple(lerax_to_gymnax_space(s) for s in space.spaces))
    else:
        raise NotImplementedError(f"Space type {type(space)} not supported")


class GymnaxEnvState(AbstractEnvState):
    env_state: gym.EnvState
    observation: Array
    reward: Float[Array, ""]
    terminal: Bool[Array, ""]
    transition_info: dict


class GymnaxToLeraxEnv(AbstractEnv[GymnaxEnvState, Array, Array]):
    """
    Wrapper of a Gymnax environment to make it compatible with Lerax.

    Note:
        For the sake of simplicity, truncation is not handled and always set to False.
        To keep the API consistent, info returned by step is always an empty dict.

    Attributes:
        action_space: Action space of the environment.
        observation_space: Observation space of the environment.
        env: Gymnax environment being wrapped.
        params: Parameters for the Gymnax environment.

    Args:
        env: Gymnax environment to wrap.
        params: Parameters for the Gymnax environment.
    """

    action_space: AbstractSpace
    observation_space: AbstractSpace

    env: gym.Environment
    params: gym.EnvParams

    def __init__(self, env: gym.Environment, params: gym.EnvParams):
        self.env = env
        self.params = params

        self.action_space = gymnax_space_to_lerax_space(env.action_space(params))
        self.observation_space = gymnax_space_to_lerax_space(
            env.observation_space(params)
        )

    def initial(self, *, key: Key) -> GymnaxEnvState:
        observation, env_state = self.env.reset_env(key, self.params)
        return GymnaxEnvState(
            env_state=env_state,
            observation=observation,
            reward=jnp.array(0.0, dtype=float),
            terminal=jnp.array(False, dtype=bool),
            transition_info={},
        )

    def transition(
        self, state: GymnaxEnvState, action: Array, *, key: Key
    ) -> GymnaxEnvState:
        observation, env_state, reward, done, _ = self.env.step_env(
            key, state.env_state, action, self.params
        )
        return GymnaxEnvState(
            env_state=env_state,
            observation=observation,
            reward=reward,
            terminal=done,
            transition_info={},
        )

    def observation(self, state: GymnaxEnvState, *, key: Key) -> Array:
        return state.observation

    def reward(
        self,
        state: GymnaxEnvState,
        action: Array,
        next_state: GymnaxEnvState,
        *,
        key: Key,
    ) -> Float[Array, ""]:
        return next_state.reward

    def terminal(self, state: GymnaxEnvState, *, key: Key) -> Bool[Array, ""]:
        return state.terminal

    def truncate(self, state: GymnaxEnvState) -> Bool[Array, ""]:
        return jnp.array(False, dtype=bool)

    def state_info(self, state: GymnaxEnvState) -> dict:
        return {}

    def transition_info(
        self, state: GymnaxEnvState, action: Array, next_state: GymnaxEnvState
    ) -> dict:
        return next_state.transition_info

    def render(self, state: GymnaxEnvState, renderer: AbstractRenderer):
        raise NotImplementedError("Rendering not implemented for GymnaxToLeraxEnv.")

    def default_renderer(self) -> AbstractRenderer:
        raise NotImplementedError(
            "Default renderer not implemented for GymnaxToLeraxEnv."
        )

    @property
    def name(self) -> str:
        return f"GymnaxToLeraxEnv({self.env.name})"


@struct.dataclass
class LeraxEnvParams(gym.EnvParams):
    pass


@struct.dataclass
class LeraxEnvState[StateType: AbstractEnvState](gym.EnvState):
    env_state: StateType
    time: Int[Array, ""]


class LeraxToGymnaxEnv[StateType: AbstractEnvState](
    gym.Environment[LeraxEnvState[StateType], LeraxEnvParams]
):
    """
    Wrapper of an Lerax environment to make it compatible with Gymnax.

    Note:
        Since Gymnax does not have a truncation concept, truncation and
        termination are combined into a single "done" signal.

    Attributes:
        env: Lerax environment being wrapped.
        state: Current state of the environment.

    Args:
        env: Lerax environment to wrap.
    """

    env: AbstractEnv[StateType, Array, Array]
    state: StateType

    def __init__(self, env: AbstractEnv[StateType, Array, Array]):
        self.env = env

    def step_env(
        self,
        key: Key,
        state: LeraxEnvState[StateType],
        action: ArrayLike,
        params: LeraxEnvParams,
    ) -> tuple[
        Array, LeraxEnvState[StateType], Float[Array, ""], Bool[Array, ""], dict
    ]:
        env_state, observation, reward, termination, truncation, info = self.env.step(
            state.env_state, jnp.asarray(action), key=key
        )
        done = termination | truncation
        return (
            observation,
            LeraxEnvState(
                env_state=env_state,
                time=state.time + 1,
            ),
            reward,
            done,
            info,
        )

    def reset_env(
        self, key: Key, params: LeraxEnvParams
    ) -> tuple[Array, LeraxEnvState[StateType]]:
        initial_key, observation_key = jr.split(key, 2)
        env_state = self.env.initial(key=initial_key)
        observation = self.env.observation(env_state, key=observation_key)

        return observation, LeraxEnvState(env_state=env_state, time=jnp.array(0))

    def get_obs(  # pyright: ignore
        self, state: LeraxEnvState[StateType], key: Key, params: LeraxEnvParams
    ) -> Array:
        return self.env.observation(state.env_state, key=key)

    def is_terminal(
        self, state: LeraxEnvState[StateType], params: LeraxEnvParams
    ) -> Bool[Array, ""]:
        return self.env.terminal(state.env_state, key=jr.key(-1))

    @property
    def name(self) -> str:
        return self.env.name

    @property
    def default_params(self) -> LeraxEnvParams:
        return LeraxEnvParams()

    def observation_space(self, params: LeraxEnvParams) -> gym_spaces.Space:
        return lerax_to_gymnax_space(self.env.observation_space)

    def action_space(self, params: LeraxEnvParams) -> gym_spaces.Space:
        return lerax_to_gymnax_space(self.env.action_space)

    def state_space(self, params: LeraxEnvParams) -> gym_spaces.Space:
        raise NotImplementedError("State space not implemented for LeraxToGymnaxEnv.")



from __future__ import annotations

import glfw
import mujoco
import numpy as np
from jax import numpy as jnp
from jaxtyping import ArrayLike, Float
from mujoco import mjx

from .base_renderer import Abstract3DRenderer


class HeadlessMujocoRenderer:
    """
    MuJoCo renderer for headless (off-screen) rendering.
    """

    model: mujoco.MjModel

    width: int
    height: int
    max_geom: int

    viewport: mujoco.MjrRect
    scene: mujoco.MjvScene
    camera: mujoco.MjvCamera
    visual_options: mujoco.MjvOption
    pert: mujoco.MjvPerturb

    markers: list[dict]
    overlays: dict[int, list[str]]
    context: mujoco.MjrContext

    def __init__(
        self,
        model: mujoco.MjModel,
        width: int = 800,
        height: int = 600,
        max_geom: int = 1000,
        visual_options: dict[int, bool] = {},
    ):
        self.width, self.height = width, height

        self.model = model
        self.markers = []
        self.overlays = {}
        self.viewport = mujoco.MjrRect(0, 0, self.width, self.height)

        self.scene = mujoco.MjvScene(self.model, max_geom)
        self.camera = mujoco.MjvCamera()
        self.camera.type = mujoco.mjtCamera.mjCAMERA_FREE
        self.camera.fixedcamid = -1
        self.camera.distance = self.model.stat.extent
        self.visual_options = mujoco.MjvOption()
        self.pert = mujoco.MjvPerturb()

        for flag, value in visual_options.items():
            self.visual_options.flags[flag] = value

        self.context = mujoco.MjrContext(
            self.model, mujoco.mjtFontScale.mjFONTSCALE_150
        )

        mujoco.mjr_setBuffer(mujoco.mjtFramebuffer.mjFB_WINDOW, self.context)

    def render(self, data: mjx.Data):
        """
        Render the current frame.

        Args:
            data: The MuJoCo simulation data to render.
        """
        for i in range(3):
            self.camera.lookat[i] = jnp.median(data.geom_xpos[:, i])

        mujoco.mjv_updateScene(
            self.model,
            mjx.get_data(self.model, data),  # pyright: ignore
            self.visual_options,
            self.pert,
            self.camera,
            mujoco.mjtCatBit.mjCAT_ALL.value,
            self.scene,
        )

        mujoco.mjr_render(self.viewport, self.scene, self.context)

    def draw(self):
        """
        Read the rendered frame into arrays.

        Returns:
            rgb_array: The RGB image array of shape (height, width, 3).
            depth_array: The depth image array of shape (height, width).
        """
        rgb_array = np.zeros((self.height, self.width, 3), dtype=np.uint8).flatten()
        depth_array = np.zeros((self.height, self.width), dtype=np.float32).flatten()

        mujoco.mjr_readPixels(rgb_array, depth_array, self.viewport, self.context)

        rgb_array = rgb_array.reshape((self.height, self.width, 3))[::-1, :]
        depth_array = depth_array.reshape((self.height, self.width))[::-1, :]

        return rgb_array, depth_array


class WindowMujocoRenderer:
    """
    On-screen MuJoCo renderer using GLFW.
    """

    model: mujoco.MjModel

    window: glfw._GLFWwindow
    width: int
    height: int
    max_geom: int

    viewport: mujoco.MjrRect
    scene: mujoco.MjvScene
    camera: mujoco.MjvCamera
    visual_options: mujoco.MjvOption
    pert: mujoco.MjvPerturb

    markers: list[dict]
    overlays: dict[int, list[str]]
    context: mujoco.MjrContext

    def __init__(
        self,
        model: mujoco.MjModel,
        width: int | None = None,
        height: int | None = None,
        max_geom: int = 1000,
        visual_options: dict[int, bool] = {},
        name: str = "MuJoCo",
    ):
        glfw.init()

        monitor_width, monitor_height = glfw.get_video_mode(
            glfw.get_primary_monitor()
        ).size

        width = monitor_width // 2 if width is None else width
        height = monitor_height // 2 if height is None else height

        glfw.window_hint(glfw.VISIBLE, glfw.TRUE)
        self.window = glfw.create_window(width, height, name, None, None)
        self.width, self.height = glfw.get_framebuffer_size(self.window)

        self.model = model
        self.markers = []
        self.overlays = {}
        self.viewport = mujoco.MjrRect(0, 0, self.width, self.height)

        self.scene = mujoco.MjvScene(self.model, max_geom)
        self.camera = mujoco.MjvCamera()
        self.camera.type = mujoco.mjtCamera.mjCAMERA_FREE
        self.camera.fixedcamid = -1
        self.camera.distance = self.model.stat.extent
        self.visual_options = mujoco.MjvOption()
        self.pert = mujoco.MjvPerturb()

        for flag, value in visual_options.items():
            self.visual_options.flags[flag] = value

        glfw.make_context_current(self.window)

        self.context = mujoco.MjrContext(
            self.model, mujoco.mjtFontScale.mjFONTSCALE_150
        )

        mujoco.mjr_setBuffer(mujoco.mjtFramebuffer.mjFB_WINDOW, self.context)

        glfw.swap_interval(1)

    def render(self, data: mjx.Data):
        """
        Render the current frame.

        Args:
            data: The MuJoCo simulation data to render.
        """
        if glfw.window_should_close(self.window):
            glfw.destroy_window(self.window)
            glfw.terminate()

        for i in range(3):
            self.camera.lookat[i] = jnp.median(data.geom_xpos[:, i])

        self.viewport.width, self.viewport.height = glfw.get_framebuffer_size(
            self.window
        )

        mujoco.mjv_updateScene(
            self.model,
            mjx.get_data(self.model, data),  # pyright: ignore
            self.visual_options,
            self.pert,
            self.camera,
            mujoco.mjtCatBit.mjCAT_ALL.value,
            self.scene,
        )

        mujoco.mjr_render(self.viewport, self.scene, self.context)

    def draw(self):
        """
        Draw the rendered frame to the window.
        """
        glfw.swap_buffers(self.window)
        glfw.poll_events()

    def close(self):
        """
        Close the rendering window.
        """
        glfw.destroy_window(self.window)
        glfw.terminate()


class MujocoRenderer(Abstract3DRenderer):
    """
    MuJoCo renderer that opens a window for on-screen rendering.

    Attributes:
        renderer: The underlying WindowMujocoRenderer instance.

    Args:
        model: The MuJoCo model to render.
        width: The width of the rendering window in pixels.
        height: The height of the rendering window in pixels.
        max_geom: The maximum number of geometries to render.
        visual_options: A dictionary of visual options for rendering.
        name: The title of the rendering window.
    """

    renderer: WindowMujocoRenderer

    def __init__(
        self,
        model: mujoco.MjModel,
        width: int | None = None,
        height: int | None = None,
        max_geom: int = 1000,
        visual_options: dict[int, bool] = {},
        name: str = "MuJoCo",
    ):
        self.renderer = WindowMujocoRenderer(
            model,
            width,
            height,
            max_geom,
            visual_options,
            name,
        )

    def open(self):
        pass

    def close(self):
        self.renderer.close()

    def render(self, data: mjx.Data):
        """
        Update the renderer with the current simulation data.

        Args:
            data: The MuJoCo simulation data to render.
        """
        self.renderer.render(data)

    def draw(self):
        """
        Update the rendering window with the latest rendered frame.
        """
        self.renderer.draw()

    def as_array(self) -> Float[ArrayLike, "H W 3"]:
        raise NotImplementedError(
            "As-array rendering is not supported in windowed mode."
        )



from __future__ import annotations

from jax import numpy as jnp
from jaxtyping import Array, Bool, Int, Key
from rich import progress, text

from lerax.env import AbstractEnvLike
from lerax.policy import AbstractPolicy
from lerax.utils import callback_with_list_wrapper, callback_wrapper

from .base_callback import (
    AbstractCallback,
    AbstractCallbackStepState,
    EmptyCallbackState,
    IterationContext,
    ResetContext,
    StepContext,
    TrainingContext,
)


def superscript_digit(digit: int) -> str:
    return "⁰¹²³⁴⁵⁶⁷⁸⁹"[digit % 10]


def superscript_int(i: int) -> str:
    return "".join(superscript_digit(int(c)) for c in str(i))


def suffixes(base: int):
    yield ""

    val = 1
    while True:
        yield f"×{base}{superscript_int(val)}"
        val += 1


def unit_and_suffix(value: float, base: int) -> tuple[float, str]:
    if base < 1:
        raise ValueError("base must be >= 1")

    unit, suffix = 1, ""
    for i, suffix in enumerate(suffixes(base)):
        unit = base**i
        if int(value) < unit * base:
            break

    return unit, suffix


class SpeedColumn(progress.ProgressColumn):
    """
    Renders human readable speed.

    https://github.com/NichtJens/rich/tree/master
    """

    def render(self, task: progress.Task) -> text.Text:
        """Show speed."""
        speed = task.finished_speed or task.speed

        if speed is None:
            return text.Text("", style="progress.percentage")
        unit, suffix = unit_and_suffix(speed, 2)
        data_speed = speed / unit
        return text.Text(f"{data_speed:.1f}{suffix} it/s", style="red")


class JITProgressBar:
    progress_bar: progress.Progress
    task: progress.TaskID

    def __init__(self, name: str, total: int | None, transient: bool = False):
        self.progress_bar = progress.Progress(
            progress.TextColumn("[progress.description]{task.description}"),
            progress.SpinnerColumn(finished_text="[green]✔"),
            progress.MofNCompleteColumn(),
            progress.BarColumn(bar_width=None),
            progress.TaskProgressColumn(),
            progress.TextColumn("["),
            progress.TimeElapsedColumn(),
            progress.TextColumn("<"),
            progress.TimeRemainingColumn(),
            progress.TextColumn("]"),
            SpeedColumn(),
            transient=transient,
        )
        self.task = self.progress_bar.add_task(f"[yellow]{name}", total=total)

    @callback_wrapper
    def start(self) -> None:
        self.progress_bar.start()

    @callback_wrapper
    def stop(self) -> None:
        self.progress_bar.stop()

    @callback_with_list_wrapper
    def update(
        self, advance: float | None = None, completed: float | None = None
    ) -> None:
        self.progress_bar.update(self.task, advance=advance, completed=completed)


class ProgressBarCallbackStepState(AbstractCallbackStepState):
    """
    State for ProgressBarCallback at each step.

    Attributes:
        steps: Number of steps taken in the current iteration.
    """

    steps: Int[Array, ""]


class ProgressBarCallback(
    AbstractCallback[EmptyCallbackState, ProgressBarCallbackStepState]
):
    """
    Callback for displaying a progress bar during training.

    Note:
        If the callback is instantiated inside a JIT-compiled function, it may
        not work correctly.

    Attributes:
        progress_bar: JITProgressBar instance for displaying progress.

    Args:
        total_timesteps: Total number of timesteps for the progress bar.
        name: Name of the progress bar. If None, a default name is generated
            from the environment and policy names.
        env: The environment being trained on. Used for naming if `name` is None.
        policy: The policy being trained. Used for naming if `name` is None.
    """

    progress_bar: JITProgressBar

    def __init__(
        self,
        total_timesteps: int | None = None,
        name: str | None = None,
        env: AbstractEnvLike | None = None,
        policy: AbstractPolicy | None = None,
    ):
        if name is None:
            if env is not None:
                if policy is not None:
                    name = f"Training {policy.name} on {env.name}"
                else:
                    name = f"Training on {env.name}"
            else:
                if policy is not None:
                    name = f"Training {policy.name}"
                else:
                    name = "Training"
        else:
            name = name

        self.progress_bar = JITProgressBar(name, total=total_timesteps)

    def reset(self, ctx: ResetContext, *, key: Key) -> EmptyCallbackState:
        self.progress_bar.start()
        return EmptyCallbackState()

    def step_reset(
        self, ctx: ResetContext, *, key: Key
    ) -> ProgressBarCallbackStepState:
        return ProgressBarCallbackStepState(jnp.array(0, dtype=int))

    def on_step(self, ctx: StepContext, *, key: Key) -> ProgressBarCallbackStepState:
        return ProgressBarCallbackStepState(ctx.state.steps + 1)

    def on_iteration(self, ctx: IterationContext, *, key: Key) -> EmptyCallbackState:
        self.progress_bar.update(completed=jnp.sum(ctx.step_state.steps))
        return ctx.state

    def on_training_start(
        self, ctx: TrainingContext, *, key: Key
    ) -> EmptyCallbackState:
        return ctx.state

    def on_training_end(self, ctx: TrainingContext, *, key: Key) -> EmptyCallbackState:
        # TODO: Fix ordered callback issue
        # self.progress_bar.stop()
        return ctx.state

    def stop(self) -> None:
        self.progress_bar.stop()

    def continue_training(self, ctx: IterationContext, *, key: Key) -> Bool[Array, ""]:
        return jnp.array(True)



from __future__ import annotations

import equinox as eqx
from distreqx import bijectors, distributions
from jax import numpy as jnp
from jaxtyping import Array, ArrayLike, Float

from .base_distribution import AbstractTransformedDistribution


class SquashedMultivariateNormalDiag(
    AbstractTransformedDistribution[Float[Array, " dims"]]
):
    """
    Multivariate Normal with squashing bijector for bounded outputs.

    Attributes:
        distribution: The underlying distreqx Transformed distribution.

    Args:
        loc: The mean of the multivariate normal distribution.
        scale_diag: The diagonal of the covariance matrix.
        high: The upper bound for bounded squashing..
        low: The lower bound for bounded squashing..
    """

    distribution: distributions.Transformed

    def __init__(
        self,
        loc: Float[ArrayLike, " dims"],
        scale_diag: Float[ArrayLike, " dims"],
        high: Float[ArrayLike, " dims"] = jnp.array(1.0),
        low: Float[ArrayLike, " dims"] = jnp.array(-1.0),
    ):
        loc = jnp.asarray(loc)
        scale_diag = jnp.asarray(scale_diag)

        (high, low) = eqx.error_if(
            (high, low),
            ~(jnp.isfinite(high) & jnp.isfinite(low)),
            "SquashedMultivariateNormalDiag requires finite low/high for all "
            "dimensions. Got non-finite bounds.",
        )

        high = jnp.broadcast_to(jnp.asarray(high), loc.shape)
        low = jnp.broadcast_to(jnp.asarray(low), loc.shape)

        mvn = distributions.MultivariateNormalDiag(loc=loc, scale_diag=scale_diag)

        sigmoid = bijectors.Sigmoid()
        affine = bijectors.ScalarAffine(scale=(high - low), shift=low)
        chain = bijectors.Chain((affine, sigmoid))
        bijector = bijectors.Block(chain, ndims=1)

        self.distribution = distributions.Transformed(mvn, bijector)

    @property
    def loc(self) -> Float[Array, " dims"]:
        assert isinstance(
            self.distribution._distribution, distributions.MultivariateNormalDiag
        )
        return self.distribution._distribution.loc

    @property
    def scale_diag(self) -> Float[Array, " dims"]:
        assert isinstance(
            self.distribution._distribution, distributions.MultivariateNormalDiag
        )
        return self.distribution._distribution.scale_diag



from __future__ import annotations

from typing import Callable

import diffrax
import equinox as eqx
from jax import lax
from jax import nn as jnn
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, Float, Key, PyTree, Shaped

from ..base_model import AbstractModelState, AbstractStatefulModel
from .term import AbstractNCDETerm, MLPNCDETerm
from .utils import safe_control

type Coeffs = tuple[
    PyTree[Shaped[Array, "times-1 ?*shape"]],
    PyTree[Shaped[Array, "times-1 ?*shape"]],
    PyTree[Shaped[Array, "times-1 ?*shape"]],
    PyTree[Shaped[Array, "times-1 ?*shape"]],
]


class NCDEState(AbstractModelState):
    ts: Float[Array, " n"]
    xs: Float[Array, " n input_size"]


class AbstractNeuralCDE[
    LatentType: Callable[[Float[Array, " out_size"]], Float[Array, " latent_size"]]
](
    AbstractStatefulModel[
        NCDEState,
        [Float[Array, ""], Float[Array, " in_size"]],
        Float[Array, " out_size"],
    ],
):
    """
    Abstract base class for Neural Controlled Differential Equations (NCDEs).

    x is used to denote the input, t is the time, z is the latent state, and y is the
    output. _s is used to denote all states of a variable (e.g. zs is the latent states
    at all times) and _1 is used to denote the last state of a variable (e.g. z1 is the
    latent state at the last time).

    The state_size defines the maximum number of inputs and corresponding states that
    are tracked. In theory the latest latent vector should contain all the important
    information about the history of inputs and states, but longer history
    allows more accurate gradients to be computed for back propagation through time.
    Inference mode can be used to disable the history and only use the latest state for
    faster computation without back propagation through time.
    """

    term: eqx.AbstractVar[AbstractNCDETerm]
    solver: eqx.AbstractVar[diffrax.AbstractSolver]

    in_size: eqx.AbstractVar[int]
    latent_size: eqx.AbstractVar[int]

    initial: eqx.AbstractVar[LatentType]
    time_in_input: eqx.AbstractVar[bool]

    history_length: eqx.AbstractVar[int]

    def solve(
        self,
        ts: Float[Array, " n"],
        xs: Float[Array, " n in_size"],
    ) -> Float[Array, " n latent_size"]:
        if isinstance(self.solver, diffrax.AbstractAdaptiveSolver):
            stepsize_controller = diffrax.PIDController(rtol=1e-3, atol=1e-6)
            dt0 = None
        else:
            stepsize_controller = diffrax.ConstantStepSize()
            dt0 = jnp.nanmean(jnp.diff(ts))

        z0 = self.z0(ts[0], xs[0])
        control = safe_control(ts, xs)
        term = diffrax.ControlTerm(self.term, control).to_ode()

        solution = diffrax.diffeqsolve(
            terms=term,
            solver=self.solver,
            t0=ts[0],
            t1=jnp.nanmax(ts),
            dt0=dt0,
            y0=z0,
            stepsize_controller=stepsize_controller,
        )

        assert solution.ys is not None
        return solution.ys[0]

    def z0(
        self, t0: Float[Array, ""], x0: Float[Array, " in_size"]
    ) -> Float[Array, " latent_size"]:
        if self.time_in_input:
            return self.initial(
                jnp.concatenate([x0, jnp.expand_dims(t0, axis=-1)], axis=-1)
            )
        else:
            return self.initial(x0)

    def next_state(
        self, state: NCDEState, ti: Float[Array, ""], xi: Float[Array, " in_size"]
    ) -> tuple[Float[Array, " num_steps"], Float[Array, " num_steps input_size"]]:
        """Add new time and input pair to the state."""
        ts, xs = state.ts, state.xs
        latest_index = jnp.nanargmax(ts)
        ts = eqx.error_if(
            ts,
            ti <= ts[latest_index],
            "new input and time must be later than all previous",
        )

        def shift() -> (
            tuple[Float[Array, " num_steps"], Float[Array, " num_steps input_size"]]
        ):
            """Shift the saved times and inputs to make room for the new pair."""
            return (
                jnp.roll(ts, -1).at[-1].set(ti),
                jnp.roll(xs, -1, axis=0).at[-1].set(xi),
            )

        def insert() -> (
            tuple[Float[Array, " num_steps"], Float[Array, " num_steps input_size"]]
        ):
            """Insert the new time and input pair at the end of the saved times and inputs."""
            return ts.at[latest_index + 1].set(ti), xs.at[latest_index + 1].set(xi)

        ts, xs = lax.cond(latest_index == self.history_length - 1, shift, insert)

        return ts, xs

    def __call__(
        self,
        state: NCDEState,
        ti: Float[Array, ""],
        xi: Float[Array, " in_size"],
    ) -> tuple[NCDEState, Float[Array, " out_size"]]:
        """Compute the next state and output given the current state and input."""
        ts, xs = self.next_state(state, ti, xi)
        zi = self.solve(ts, xs)
        return NCDEState(ts, xs), zi

    def reset(self) -> NCDEState:
        """Reset the state to an empty state."""
        times = jnp.full((self.history_length,), jnp.nan, dtype=float)
        inputs = jnp.full((self.history_length, self.in_size), jnp.nan, dtype=float)
        return NCDEState(times, inputs)


class MLPNeuralCDE(AbstractNeuralCDE):
    term: MLPNCDETerm
    solver: diffrax.AbstractSolver

    initial: eqx.nn.MLP
    time_in_input: bool

    in_size: int
    latent_size: int

    history_length: int

    def __init__(
        self,
        in_size: int,
        latent_size: int,
        *,
        term: MLPNCDETerm | None = None,
        initial: eqx.nn.MLP | None = None,
        field_width: int = 64,
        field_depth: int = 2,
        field_activation: Callable[
            [Float[Array, " width"]], Float[Array, " width"]
        ] = jnn.softplus,
        field_final_activation: Callable[
            [Float[Array, " width"]], Float[Array, " width"]
        ] = jnn.tanh,
        initial_width: int = 64,
        initial_depth: int = 1,
        initial_state_activation: Callable[
            [Float[Array, " width"]], Float[Array, " width"]
        ] = jnn.relu,
        initial_state_final_activation: Callable[
            [Float[Array, " width"]], Float[Array, " width"]
        ] = lambda x: x,
        solver: diffrax.AbstractSolver | None = None,
        time_in_input: bool = False,
        history_length: int = 16,
        key: Key,
    ):
        term_key, initial_key = jr.split(key, 2)

        self.solver = solver or diffrax.Tsit5()
        self.time_in_input = time_in_input
        self.history_length = history_length

        self.in_size = in_size
        self.latent_size = latent_size

        self.term = (
            term
            if term is not None
            else MLPNCDETerm(
                input_size=in_size,
                state_size=latent_size,
                width_size=field_width,
                depth=field_depth,
                key=term_key,
                add_time=time_in_input,
                activation=field_activation,
                final_activation=field_final_activation,
            )
        )

        self.initial = (
            initial
            if initial is not None
            else eqx.nn.MLP(
                in_size=in_size + int(time_in_input),
                out_size=latent_size,
                width_size=initial_width,
                depth=initial_depth,
                key=initial_key,
                activation=initial_state_activation,
                final_activation=initial_state_final_activation,
            )
        )
